#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


#
# Global settings
#

# Should the driver check that A10 appliances are alive before
# successfullAy initializing?

# verify_appliances = False

# Should the driver store some meta-info in a database?
# Needed for tenant<->appliance persistence if the number of appliances
# is changed for any reason. Setting this to true means that you will
# also need to run "a10-manage upgrade" after installing or upgrading
# the a10-neutron-lbaas package.
# Recommended for all new installs.

# use_database = False

# The SQLAlchemy connection string to use to connect to the database.
# If None, and use_database is True, the driver will attempt to use
# the configured neutron database.

# database_connection = None

# Sometimes we need things from neutron. We will look in the usual places,
# but this is here if you need to override the location.

# neutron_conf_dir = '/etc/neutron'

# If True, use the OpenStack lbaas member object UUID as the server name
# in ACOS. Recommended for all new installs using ACOS 4.0.0 or greater.

# member_name_use_uuid = False

# If not None, use this keystone auth URL instead of the one from the
# neutron.conf file.

# keystone_auth_url = None

# Which version of the keystone protocol to use

# keystone_version = 2

# Certain functions of this driver can be overridden by passing in an alternate
# set of plumbing hooks, including scheduling where a tenant's VIPs are going
# to be created, and what to do after object creation.
#
# The default hooks class is the first, PlumbingHooks, which is meant for
# hardware appliances and manually orchestrated A10 devices. This hook
# primarily makes use of the 'devices' hash table below to find available
# ACOS devices.
#
# The second hooks class provided is the VThunderPlumbingHooks, which will
# spawn a virtual appliance for every VIP, inside a service tenant space.
# It's settings primarily come from the 'vthunder' hash defined at the
# bottom of this file.
#
# More advanced hooks/schedulers can be added by the operator or by
# A10 professional services.

# import a10_neutron_lbaas.plumbing.simple as hooks
# plumbing_hooks_class = hooks.PlumbingHooks

# import a10_neutron_lbaas.plumbing.vthunder_per_vip as hooks
# plumbing_hooks_class = hooks.VThunderPerVIPPlumbingHooks

# Nova API version; defaults to '2.1' (hint: use '2' for kilo.)

# nova_api_version = '2.1'

#
# Main devices dictionary, containing a list of available ACOS devices.
#

devices = {
    # A sample ACOS 2.7.2 box
    # "ax2": {
    #     "host": "10.10.100.20",
    #     "conn-limit": 8000000,
    #     "port": 8443,
    #     "username": "admin",
    #     "password": "a10",
    #     "api_version": "2.1",
    # },
    # A sample ACOS 4.0.1 box
    # "ax4": {
    #     "host": "10.10.100.20",
    #     "conn-limit": 8000000,
    #     "port": 443,
    #     "username": "admin",
    #     "password": "a10",
    #     "api_version": "3.0",
    # },
    # The complete list of available options for a device entry, with
    # their default values.
    # "axN": {
    #
    # Hostname of ACOS device
    #     "host": <required>,
    #
    # Protocol and port number for AxAPI on ACOS device
    #     "protocol": "https",
    #     "port": 443,
    #
    # Specify the maximum number of concurrent connections allowed on a
    # real server.
    #    "conn-limit": 8000000,
    #
    # Admin username for ACOS device
    #     "username": <required>,
    #
    # Admin password for ACOS device
    #     "password": <required>,
    #
    # Which version of AxAPI to use; "2.1" for ACOS<3.0, and "3.0" for >=3.0
    #     "api_version": "2.1",
    #
    # status of device; True if online and ready for use
    #     "status": True,
    #
    # Set to true if you want automatic source nat on vports
    #     "autosnat": False,
    #
    # Set to the name of a nat pool to use that pool for source nat on vports
    # The nat pool must already exist on the ACOS device.
    #     "source_nat_pool": None,
    #
    # Partition method; "LSI" to put all slb's in a single shared partition,
    # or "ADP" to use a partition per tenant. Partitions are RBA style in
    # ACOS 2.x, and L3V in ACOS 4.x.
    #     "v_method": "LSI",
    #
    # If using a shared partition (v_method=LSI), then this field configures
    # which partition to use. By default, it is the main shared partition.
    #     "shared_partition": "shared",
    #
    # For nova based pool/service-group members, setting this to True will
    # cause the driver to lookup the nova server's floating ip and use that
    # for communication instead of its neutron port IP address.
    #     "use_float": False,
    #
    # Virtual servers will be created on this VRID. The VRID must already
    # be configured on the device. Example values: None, 1, 2, ...
    #     "default_virtual_server_vrid": None,
    #
    # Enable IP in IP on vports.
    #     "ipinip": False,
    #
    # Contains a list of hostnames or IP addresses that the driver will run
    # the 'ha sync' command against whenever a write operation occurs.
    #     "ha_sync_list": [],
    #
    # Enable or disable calling write memory directly after any operation that
    # changes ACOS's running state. Turning this off also disables all ha sync
    # operations, regardless of the settings in ha_sync_list.
    #     "write_memory": True,
    # },
}

# vthunder = {
#     'username': 'admin',
#     'password': 'a10',

#     'nova_flavor': 'vthunder.small',  # 1 core, 4096MB ram, 12GB disk
#     'glance_image': None,

#     'vthunder_management_network': 'private-mgmt',
#     'vthunder_data_networks': [ 'vip-net', 'member-net' ],

# # License the launched instances
#
#     'license_manager': {
#              "hosts": [
#                     {"ip": "pdx.a10cloud.com", "port": 443},
#                     {"ip": "sfo.a10cloud.com", "port": 443},
#                     {"ip": "iad.a10cloud.com", "port": 443}
#             ],
#             "serial": "SNxxxxxxxxxxxxxxxx",
#             "instance-name": "openstack_instance",
#             "bandwidth-base": 100,
#             "interval": 3,
#             "use-mgmt-port": True
#     },

# # Configure dns resolver
#
#     'dns_resolver': {
#         'primary': '192.0.2.4',
#         'secondary': '192.0.2.5',
#     },

# # Configure statistics collection
#
#     'sflow_collector': {
#          'host': '10.20.100.7',
#          'port': 6343
#     }

#     # The following are required if you wish to use a service tenant.
#     'service_tenant': {
#         'tenant_name': 'a10-service',
#         'username': '',
#         'password': ''
#     }
# }
vport_defaults = {}



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class FakeModel(dict, object):
    def __getitem__(self, key, default=None):
        attr = getattr(self, key, default)
        return attr

    def get(self, key, default=None):
        return getattr(self, key, default)


class FakeVIP(FakeModel):
    def __init__(self, pers="", vip_id="id1"):
        super(FakeVIP, self).__init__()
        self.tenant_id = 'ten1'
        self.id = vip_id
        self.protocol = 'HTTP'
        self.admin_state_up = True
        self.address = '1.1.1.1'
        self.protocol_port = '80'
        self.pool_id = 'pool1'
        self.subnet_id = 'subnet1'
        self.root_loadbalancer = self
        if pers:
            self.session_persistence = {'type': pers}


class FakeHM(FakeModel):
    def __init__(self, type=None):
        super(FakeHM, self).__init__()
        self.id = "hm01"
        self.name = "hm01"
        self.tenant_id = 'tenv1'
        self.type = type
        self.delay = '5'
        self.timeout = 5
        self.max_retries = '5'
        self.pools = []
        self.root_loadbalancer = FakeVIP()
        if type in ['HTTP', 'HTTPS']:
            self.http_method = 'GET'
            self.url_path = '/'
            self.expected_codes = '200'

    def copy(self):
        return self


class FakePool(FakeModel):
    def __init__(self, protocol='TCP', method='ROUND_ROBIN'):
        super(FakePool, self).__init__()
        self.id = 'id1'
        self.tenant_id = 'ten1'
        self.protocol = protocol
        self.lb_method = method
        self.members = []
        self.root_loadbalancer = FakeVIP()


class FakeMember(FakeModel):
    def __init__(self, admin_state_up=True, pool_id='pool1'):
        super(FakeMember, self).__init__()
        self.address = '1.1.1.1'
        self.id = 'id1'
        self.tenant_id = 'ten1'
        self.admin_state_up = admin_state_up
        self.pool_id = pool_id
        self.protocol_port = '80'
        self.root_loadbalancer = FakeVIP()



# Copyright 2016, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Added status column to bindings table

Revision ID: 4ffd7ee0f175
Revises: e3d5134c59a7
Create Date: 2016-12-14 01:01:28.281713

"""

# revision identifiers, used by Alembic.
revision = '4ffd7ee0f175'
down_revision = 'e3d5134c59a7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('a10_certificatelistenerbindings',
                  sa.Column('status', sa.Integer, nullable=False, default=0, server_default='0')
                  )


def downgrade():
    op.drop_column('a10_certificatelistenerbindings', 'status')



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas.a10_context as a10_context


class A10Context(a10_context.A10Context):
    pass


class A10WriteContext(a10_context.A10WriteContext):
    pass


class A10WriteStatusContext(a10_context.A10WriteContext):

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.openstack_driver._active(
                self.openstack_context,
                self.handler._model_type(),
                self.openstack_lbaas_obj['id'])
        else:
            self.openstack_driver._failed(
                self.openstack_context,
                self.handler._model_type(),
                self.openstack_lbaas_obj['id'])

        super(A10WriteStatusContext, self).__exit__(exc_type, exc_value,
                                                    traceback)


class A10DeleteContext(a10_context.A10DeleteContextBase):

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.openstack_driver._db_delete(
                self.openstack_context,
                self.handler._model_type(),
                self.openstack_lbaas_obj['id'])

        super(A10DeleteContext, self).__exit__(exc_type, exc_value, traceback)

    def remaining_root_objects(self):
        ctx = self.openstack_context
        d = self.handler.openstack_driver
        n = d._pool_total(ctx, self.tenant_id)
        return n


class A10WriteHMStatusContext(a10_context.A10WriteContext):

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.openstack_driver._hm_active(
                self.openstack_context,
                self.openstack_lbaas_obj['id'],
                self.openstack_lbaas_obj['pool_id'])
        else:
            self.openstack_driver._hm_failed(
                self.openstack_context,
                self.openstack_lbaas_obj['id'],
                self.openstack_lbaas_obj['pool_id'])

        super(A10WriteHMStatusContext, self).__exit__(exc_type, exc_value,
                                                      traceback)


class A10DeleteHMContext(a10_context.A10WriteContext):

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            self.openstack_driver._hm_db_delete(
                self.openstack_context,
                self.openstack_lbaas_obj['id'],
                self.openstack_lbaas_obj['pool_id'])

        super(A10DeleteHMContext, self).__exit__(exc_type, exc_value,
                                                 traceback)



#   Copyright 2014-2016,  A10 Networks
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa
import sqlalchemy.orm as orm

from a10_neutron_lbaas.db import model_base


class Certificate(model_base.A10BaseMixin, model_base.A10Base):
    __tablename__ = "a10_certificates"
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text(1024), nullable=True)
    cert_data = sa.Column(sa.Text(8000), nullable=False)
    key_data = sa.Column(sa.Text(8000), nullable=False)
    intermediate_data = sa.Column(sa.Text(8000), nullable=True)
    password = sa.Column(sa.String(1024), nullable=True)


class CertificateListenerBinding(model_base.A10BaseMixin, model_base.A10Base):
    __tablename__ = "a10_certificatelistenerbindings"
    certificate_id = sa.Column(sa.String(36), sa.ForeignKey("a10_certificates.id"),
                               nullable=False)
    certificate = orm.relationship(Certificate, uselist=False)
    listener_id = sa.Column(sa.String(36), nullable=False)
    # This is a TINYINT/BYTE field depending on SQL implementation
    status = sa.Column(sa.Integer(), default=0, nullable=False, server_default="0")



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ConfigParser as ini
import logging
import os
import runpy
import sys

from debtcollector import removals

from a10_neutron_lbaas import a10_exceptions as a10_ex
from a10_neutron_lbaas.etc import config as blank_config
from a10_neutron_lbaas.etc import defaults

LOG = logging.getLogger(__name__)


class ConfigModule(object):
    def __init__(self, d, provider=None):
        self.__dict__.update(d)

        if provider is None or 'providers' not in d or provider not in d['providers']:
            return None

        for k, v in d['providers'][provider].items():
            if isinstance(v, dict):
                if k not in self.__dict__:
                    self.__dict__[k] = {}
                self.__dict__[k].update(v)
            else:
                self.__dict__[k] = v

    @classmethod
    def load(cls, path, provider=None):
        d = runpy.run_path(path)
        return ConfigModule(d, provider=provider)


class A10Config(object):

    def __init__(self, config_dir=None, config=None, provider=None):
        if config is not None:
            self._config = config
            self._load_config()
            return

        self._config_dir = self._find_config_dir(config_dir)
        self._config_path = os.path.join(self._config_dir, "config.py")

        try:
            self._config = ConfigModule.load(self._config_path, provider=provider)
        except IOError:
            LOG.error("A10Config could not find %s", self._config_path)
            self._config = blank_config

        self._load_config()

    def _find_config_dir(self, config_dir):
        # Look for config in the virtual environment
        # virtualenv puts the original prefix in sys.real_prefix
        # pyenv puts it in sys.base_prefix
        venv_d = os.path.join(sys.prefix, 'etc/a10')
        has_prefix = (hasattr(sys, 'real_prefix') or hasattr(sys, 'base_prefix'))

        env_override = os.environ.get('A10_CONFIG_DIR', None)
        if config_dir is not None:
            d = config_dir
        elif env_override is not None:
            d = env_override
        elif has_prefix and os.path.exists(venv_d):
            d = venv_d
        elif os.path.exists('/etc/neutron/services/loadbalancer/a10networks'):
            d = '/etc/neutron/services/loadbalancer/a10networks'
        else:
            d = '/etc/a10'

        return d

    def _load_config(self):
        # Global defaults
        for dk, dv in defaults.GLOBAL_DEFAULTS.items():
            if not hasattr(self._config, dk):
                LOG.debug("setting global default %s=%s", dk, dv)
                setattr(self._config, dk, dv)
            else:
                LOG.debug("global setting %s=%s", dk, getattr(self._config, dk))

        self._devices = {}
        for k, v in self._config.devices.items():
            if 'status' in v and not v['status']:
                LOG.debug("status is False, skipping dev: %s", v)
            else:
                for x in defaults.DEVICE_REQUIRED_FIELDS:
                    if x not in v:
                        msg = "device %s missing required value %s, skipping" % (k, x)
                        LOG.error(msg)
                        raise a10_ex.InvalidDeviceConfig(msg)

                v['key'] = k
                self._devices[k] = v

                # Old configs had a name field
                if 'name' not in v:
                    self._devices[k]['name'] = k

                # Figure out port and protocol
                protocol = self._devices[k].get('protocol', 'https')
                port = self._devices[k].get(
                    'port', {'http': 80, 'https': 443}[protocol])
                self._devices[k]['protocol'] = protocol
                self._devices[k]['port'] = port

                # Device defaults
                for dk, dv in defaults.DEVICE_OPTIONAL_DEFAULTS.items():
                    if dk not in self._devices[k]:
                        self._devices[k][dk] = dv

                LOG.debug("A10Config, device %s=%s", k, self._devices[k])

        self._vthunder = None

        if hasattr(self._config, 'vthunder'):
            self._vthunder = self._config.vthunder

            for x in defaults.VTHUNDER_REQUIRED_FIELDS:
                if x not in self._vthunder:
                    msg = "vthunder definition missing required value %s, skipping" % x
                    LOG.error(msg)
                    raise a10_ex.InvalidDeviceConfig(msg)

            # vThunder defaults
            for vk, vv in defaults.VTHUNDER_OPTIONAL_DEFAULTS.items():
                if vk not in self._vthunder:
                    self._vthunder[vk] = vv
            for dk, dv in defaults.DEVICE_OPTIONAL_DEFAULTS.items():
                if dk not in self._vthunder:
                    self._vthunder[dk] = dv

        # Setup db foo
        if self._config.use_database and self._config.database_connection is None:
            self._config.database_connection = self._get_neutron_db_string()

        if self._config.keystone_auth_url is None:
            self._config.keystone_auth_url = self._get_neutron_conf(
                'keystone_authtoken', 'auth_uri')

        # TODO(mdurrant) - There's a way to do this with getattr/setattr
        self._vport_defaults = {}
        if hasattr(self._config, "vport_defaults"):
            self._vport_defaults = self._config.vport_defaults

        self._vport_expressions = {}
        if hasattr(self._config, "vport_expressions"):
            self._vport_expressions = self._config.vport_expressions

        self._virtual_server_expressions = {}
        if hasattr(self._config, "virtual_server_expressions"):
            self._virtual_server_expressions = self._config.virtual_server_expressions

        self._service_group_expressions = {}
        if hasattr(self._config, "service_group_expressions"):
            self._service_group_expressions = self._config.service_group_expressions

        self._member_expressions = {}
        if hasattr(self._config, "member_expressions"):
            self._member_expressions = self._config.member_expressions

        self._monitor_expressions = {}
        if hasattr(self._config, "monitor_expressions"):
            self._monitor_expressions = self._config.monitor_expressions

        # Setup some backwards compat stuff
        self.config = OldConfig(self)

    # We don't use oslo.config here, in a weak attempt to avoid pulling in all
    # the many openstack dependencies. If this proves problematic, we should
    # shoot this manual parser in the head and just use the global config
    # object.
    def _get_neutron_conf(self, section, option):
        neutron_conf_dir = os.environ.get('NEUTRON_CONF_DIR', self._config.neutron_conf_dir)
        neutron_conf = '%s/neutron.conf' % neutron_conf_dir

        if os.path.exists(neutron_conf):
            LOG.debug("found neutron.conf file in /etc")
            n = ini.ConfigParser()
            n.read(neutron_conf)
            try:
                return n.get(section, option)
            except (ini.NoSectionError, ini.NoOptionError):
                pass

    def _get_neutron_db_string(self):
        z = self._get_neutron_conf('database', 'connection')

        if z is None:
            raise a10_ex.NoDatabaseURL('must set db connection url or neutron dir in config.py')

        LOG.debug("using %s as db connect string", z)
        return z

    def get(self, key):
        return getattr(self._config, key)

    def get_device(self, device_name, db_session=None):
        if device_name in self._devices:
            return self._devices.get(device_name, {})
        if self.get('use_database'):
            from a10_neutron_lbaas.db import models

            instance = models.A10DeviceInstance.find_by(name=device_name, db_session=db_session)
            if instance is not None:
                self._devices[device_name] = instance.as_dict()
                return self._devices[device_name]
        return None

    # TODO(dougwig) -- later - use of this method should be considered a scalability killer
    def get_devices(self, db_session=None):
        if self.get('use_database'):
            from a10_neutron_lbaas.db import models

            d = dict(self._devices.items())
            for x in models.A10DeviceInstance.find_all():
                d[x.name] = x.as_dict()
            return d
        return self._devices

    def get_vthunder_config(self):
        return self._vthunder

    def get_vport_defaults(self):
        return self._vport_defaults

    def get_vport_expressions(self):
        return self._vport_expressions

    def get_virtual_server_expressions(self):
        return self._virtual_server_expressions

    def get_service_group_expressions(self):
        return self._service_group_expressions

    def get_member_expressions(self):
        return self._member_expressions

    def get_monitor_expressions(self):
        return self._monitor_expressions

    # backwards compat
    @removals.remove
    @property
    def devices(self):
        return self.config.devices

    @removals.remove
    @property
    def use_database(self):
        return self.config.use_database

    @removals.remove
    @property
    def database_connection(self):
        return self.config.database_connection

    @removals.remove
    @property
    def verify_appliances(self):
        return self.config.verify_appliances


# backwards compat
class OldConfig(object):

    def __init__(self, main_config):
        self._config = main_config

    @removals.remove
    @property
    def devices(self):
        return self._config.get_devices()

    @removals.remove
    @property
    def use_database(self):
        return self._config.get('use_database')

    @removals.remove
    @property
    def database_connection(self):
        return self._config.get('database_connection')

    @removals.remove
    @property
    def verify_appliances(self):
        return self._config.get('verify_appliances')



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import test_base


class TestA10Config(test_base.UnitTestBase):

    def test_verify_appliances(self):
        self.assertTrue(self.a.config.get('verify_appliances'))

    def test_num_appliances(self):
        # Everytime we update the test config, this test has to be updated
        # A better test would seem to be be parsnig the JSON structure found in the file
        # and comparing that against what we get in devices.
        # This actually tests the number of devices with status == True
        self.assertEqual(10, len(self.a.config.get_devices()))

    def test_expected_ports(self):
        self.assertEqual(8443, self.a.config.get_device('ax1')['port'])
        self.assertEqual(80, self.a.config.get_device('ax3')['port'])
        self.assertEqual(443, self.a.config.get_device('ax4')['port'])

    def test_expected_protocols(self):
        self.assertEqual('https', self.a.config.get_device('ax1')['protocol'])
        self.assertEqual('http', self.a.config.get_device('ax3')['protocol'])
        self.assertEqual('https', self.a.config.get_device('ax4')['protocol'])

    def test_v_method(self):
        for k, v in self.a.config.get_devices().items():
            self.assertEqual('LSI', v['v_method'].upper())

    def test_alternate_shared_partition(self):
        self.assertTrue(self.a.config.get_device('axadp-alt')['shared_partition'])

    def test_ip_in_ip(self):
        expected = True
        actual = False
        for k, v in self.a.config.get_devices().items():
            if "ip_in_ip" in v:
                actual = v['ip_in_ip']
                self.assertEqual(expected, actual)

    # TODO(dougwig) -- test new a10_config members
    # def test_image_defaults(self):
    #     self.assertIsNotNone(self.a.config.image_defaults)

    # def test_image_defaults_members(self):
    #     image_defaults = self.a.config.image_defaults
    #     actual = image_defaults.keys()
    #     expected = ["name", "id", "visibility", "tags", "min_disk",
    #                 "min_ram", "container_format", "protected",
    #                 "properties", "disk_format"]
    #     self.assertListEqual(sorted(expected), sorted(actual))

    # def test_instance_defaults(self):
    #     self.assertIsNotNone(self.a.config.instance_defaults)

    def test_backwards_compat(self):
        self.assertEqual(self.a.config.get_devices(), self.a.config.devices)
        self.assertEqual(self.a.config.get_devices(), self.a.config.config.devices)
        self.assertEqual(self.a.config.get(
            'database_connection'), self.a.config.database_connection)
        self.assertEqual(self.a.config.get('use_database'), self.a.config.use_database)
        self.assertEqual(self.a.config.get('verify_appliances'), self.a.config.verify_appliances)
        self.assertEqual(self.a.config.get(
            'database_connection'), self.a.config.config.database_connection)
        self.assertEqual(self.a.config.get('use_database'), self.a.config.config.use_database)
        self.assertEqual(self.a.config.get(
            'verify_appliances'), self.a.config.config.verify_appliances)
        self.assertEqual(self.a.config.get(
            'vport_defaults'), self.a.config.get_vport_defaults())


class TestA10ConfigProvider(test_base.UnitTestBase):

    def setUp(self):
        super(TestA10ConfigProvider, self).setUp({'provider': 'prov1'})

    def test_top_level(self):
        self.assertEqual(self.a.config.get('who_should_win'), 'the-doctor')
        self.assertEqual(self.a.config.get('best_spaceship'), 'tardis')

    def test_vthunder_api_version(self):
        v = self.a.config.get_vthunder_config()
        self.assertEqual(v['api_version'], '9.9')
        self.assertEqual(v['nova_flavor'], 'acos.min')




# Copyright 2014, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def hm_type(c, os_hm_type):
    hm_map = {
        'PING': c.client.slb.hm.ICMP,
        'TCP': c.client.slb.hm.TCP,
        'HTTP': c.client.slb.hm.HTTP,
        'HTTPS': c.client.slb.hm.HTTPS
    }
    return hm_map[os_hm_type]


def service_group_lb_method(c, os_method):
    z = c.client.slb.service_group
    lb_methods = {
        'ROUND_ROBIN': z.ROUND_ROBIN,
        'LEAST_CONNECTIONS': z.LEAST_CONNECTION,
        'SOURCE_IP': z.SOURCE_IP_HASH,
        'WEIGHTED_ROUND_ROBIN': z.WEIGHTED_ROUND_ROBIN,
        'WEIGHTED_LEAST_CONNECTION': z.WEIGHTED_LEAST_CONNECTION,
        'LEAST_CONNECTION_ON_SERVICE_PORT':
            z.LEAST_CONNECTION_ON_SERVICE_PORT,
        'WEIGHTED_LEAST_CONNECTION_ON_SERVICE_PORT':
            z.WEIGHTED_LEAST_CONNECTION_ON_SERVICE_PORT,
        'FAST_RESPONSE_TIME': z.FAST_RESPONSE_TIME,
        'LEAST_REQUEST': z.LEAST_REQUEST,
        'STRICT_ROUND_ROBIN': z.STRICT_ROUND_ROBIN,
        'STATELESS_SOURCE_IP_HASH': z.STATELESS_SOURCE_IP_HASH,
        'STATELESS_DESTINATION_IP_HASH': z.STATELESS_DESTINATION_IP_HASH,
        'STATELESS_SOURCE_DESTINATION_IP_HASH':
            z.STATELESS_SOURCE_DESTINATION_IP_HASH,
        'STATELESS_PER_PACKET_ROUND_ROBIN':
            z.STATELESS_PER_PACKET_ROUND_ROBIN,
    }
    return lb_methods[os_method]


def service_group_protocol(c, os_protocol):
    z = c.client.slb.service_group
    protocols = {
        'HTTP': z.TCP,
        'HTTPS': z.TCP,
        'TERMINATED_HTTPS': z.TCP,
        'TCP': z.TCP,
        'UDP': z.UDP
    }
    return protocols[os_protocol]


def vip_protocols(c, os_protocol):
    z = c.client.slb.virtual_server.vport
    protocols = {
        'TCP': z.TCP,
        'UDP': z.UDP,
        'HTTP': z.HTTP,
        'HTTPS': z.TCP,
        'TERMINATED_HTTPS': z.HTTPS,
        'OTHERS': z.OTHERS,
        'RTSP': z.RTSP,
        'FTP': z.FTP,
        'MMS': z.MMS,
        'SIP': z.SIP,
        'FAST_HTTP': z.FAST_HTTP,
        'GENERIC_PROXY': z.GENERIC_PROXY,
        'SSL_PROXY': z.SSL_PROXY,
        'SMTP': z.SMTP,
        'SIP_TCP': z.SIP_TCP,
        'SIPS': z.SIPS,
        'DIAMETER': z.DIAMETER,
        'DNS_UDP': z.DNS_UDP,
        'TFTP': z.TFTP,
        'DNS_TCP': z.DNS_TCP,
        'RADIUS': z.RADIUS,
        'MYSQL': z.MYSQL,
        'MSSQL': z.MSSQL,
        'FIX': z.FIX,
        'SMPP_TCP': z.SMPP_TCP,
        'SPDY': z.SPDY,
        'SPDYS': z.SPDYS,
        'FTP_PROXY': z.FTP_PROXY
    }
    return protocols[os_protocol]



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
import handler_base_v2
import logging
import v2_context as a10

from a10_neutron_lbaas.acos import openstack_mappings

LOG = logging.getLogger(__name__)


class HealthMonitorHandler(handler_base_v2.HandlerBaseV2):

    def _name(self, hm):
        return hm.id[0:28]

    def _set(self, c, set_method, context, hm, **kwargs):
        hm_name = self._meta_name(hm)
        method = None
        url = None
        expect_code = None
        os_name = hm.name
        if hm.type in ['HTTP', 'HTTPS']:
            method = hm.http_method
            url = hm.url_path
            expect_code = hm.expected_codes

        self._ensure_timeout_does_not_exceed_delay(hm)

        args = self.meta(hm, 'hm', {})

        set_method(hm_name, openstack_mappings.hm_type(c, hm.type),
                   hm.delay, hm.timeout, hm.max_retries,
                   method=method, url=url, expect_code=expect_code,
                   config_defaults=self._get_config_defaults(c, os_name),
                   axapi_args=args)

    def _ensure_timeout_does_not_exceed_delay(self, hm):
        if hm.delay < hm.timeout and hm.timeout > 0:
            hm.delay = hm.timeout

    def _disable_health_monitor(self, c, context, hm):
        c.client.slb.service_group.update(
            self._pool_name(context, pool=hm.pool),
            health_monitor="", health_check_disable=True)

    def _create(self, c, context, hm, **kwargs):
        try:
            self._set(c, c.client.slb.hm.create, context, hm, **kwargs)
        except acos_errors.Exists:
            pass

        # Disable any potentially existing health monitor.
        c.client.slb.service_group.update(
            self._pool_name(context, pool=hm.pool),
            health_monitor="", health_check_disable=True)

        c.client.slb.service_group.update(
            self._pool_name(context, pool=hm.pool),
            health_monitor=self._meta_name(hm), health_check_disable=False)

    def create(self, context, hm, **kwargs):
        LOG.debug("HealthMonitorHandler.create(): hm=%s, context=%s" % (dir(hm), context))
        with a10.A10WriteStatusContext(self, context, hm) as c:
            self._create(c, context, hm, **kwargs)

    def update(self, context, old_hm, hm):
        with a10.A10WriteStatusContext(self, context, hm) as c:
            if old_hm.pool and not hm.pool:
                pool_name = self._pool_name(context, pool=old_hm.pool)
                c.client.slb.service_group.update(pool_name,
                                                  health_monitor="",
                                                  health_check_disable=True)
            elif old_hm.pool != hm.pool:
                pool_name = self._pool_name(context, pool=hm.pool)

                # Remove any existing association.  This should be moved into a method.
                c.client.slb.service_group.update(pool_name,
                                                  health_monitor="",
                                                  health_check_disable=True)
                c.client.slb.service_group.update(pool_name,
                                                  health_monitor=self._meta_name(hm),
                                                  health_check_disable=False)
            self._set(c, c.client.slb.hm.update, context, hm)

    def _delete(self, c, context, hm):
        LOG.debug("HealthMonitorHandler.delete(): hm=%s, context=%s" % (hm, context))
        pool_name = self._pool_name(context, pool=hm.pool)
        LOG.debug("HealthMonitorHandler.delete(): Updating Pool %s" % (pool_name))
        c.client.slb.service_group.update(pool_name, health_monitor="",
                                          health_check_disable=True)
        c.client.slb.hm.delete(self._meta_name(hm))

    def delete(self, context, hm):
        with a10.A10DeleteContext(self, context, hm) as c:
            self._delete(c, context, hm)

    def _get_expressions(self, c):
        rv = {}
        rv = c.a10_driver.config.get_monitor_expressions()
        return rv



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Nova instance ID not required

Revision ID: bc5626a5af2a
Revises: 3c7123f2aeba
Create Date: 2016-10-27 18:47:41.163902

"""

# revision identifiers, used by Alembic.
revision = 'bc5626a5af2a'
down_revision = '3c7123f2aeba'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column('a10_device_instances',
                    sa.Column('nova_instance_id', sa.String(36), nullable=True))


def downgrade():
    op.alter_column('a10_device_instances',
                    sa.Column('nova_instance_id', sa.String(36), nullable=False))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


LB_METHOD_ROUND_ROBIN = 'ROUND_ROBIN'
PROTOCOL_TERMINATED_HTTPS = 'TERMINATED_HTTPS'
PROTOCOL_HTTP = 'HTTP'
PROTOCOL_TCP = 'TCP'
PROTOCOL_HTTPS = 'HTTPS'

# Statuses for a10 objects (certificate bindings)
STATUS_CREATING = 0
STATUS_CREATED = 1
STATUS_DELETING = 2



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Please refer to a10_neutron_lbaas.etc.config for documentation
# about these settings.

import a10_neutron_lbaas.plumbing_hooks


GLOBAL_DEFAULTS = {
    "verify_appliances": False,
    "use_database": False,
    "database_connection": None,
    "neutron_conf_dir": '/etc/neutron',
    "member_name_use_uuid": False,
    "keystone_auth_url": None,
    "keystone_version": 2,
    "plumbing_hooks_class": a10_neutron_lbaas.plumbing_hooks.PlumbingHooks,
    "nova_api_version": "2.1",
    "vport_defaults": {},
}

DEVICE_REQUIRED_FIELDS = [
    "host",
    "username",
    "password",
]

DEVICE_OPTIONAL_DEFAULTS = {
    "protocol": "https",
    "port": 443,
    "api_version": "2.1",
    "status": True,
    "autosnat": False,
    "v_method": "LSI",
    "shared_partition": "shared",
    "use_float": False,
    "default_virtual_server_vrid": None,
    "ipinip": False,
    "ha_sync_list": [],
    "write_memory": True,
    "vport_defaults": {},
    # "max_instance": 5000,
    # "method": "hash",

    # "max_partitions": 10,
    # "per_partition_lif_max": 10,
    # "peer_mgmt_ip": "",
    # "peer_mgmt_port": "",
    # "default_virtual_server_vrid": 1,
    # "vlan": 0,
    # "gateway_mode": 1,
}

VTHUNDER_REQUIRED_FIELDS = [
    'username',
    'password',

    'nova_flavor',
    'glance_image',

    'vthunder_management_network',
    'vthunder_data_networks',
]

VTHUNDER_OPTIONAL_DEFAULTS = DEVICE_OPTIONAL_DEFAULTS
VTHUNDER_OPTIONAL_DEFAULTS = dict(DEVICE_OPTIONAL_DEFAULTS)
VTHUNDER_OPTIONAL_DEFAULTS.update({
    'autosnat': True,
    'v_method': 'LSI',
    'api_version': '3.0',
})



# Copyright (C) 2016, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import logging
import six

import a10_openstack_lib.resources.a10_certificate as a10_certificate
import a10_openstack_lib.resources.validators as a10_validators

from neutron.api import extensions as nextensions
from neutron.api.v2 import resource_helper

# neutron.services got moved to neutron_lib
try:
    # F811 (redefinition of ServicePluginBase) suppressed
    from neutron.services.service_base import ServicePluginBase  # noqa
except (ImportError, AttributeError):
    pass

try:
    # F811 (redefinition of ServicePluginBase) suppressed
    from neutron_lib.services.base import ServicePluginBase  # noqa
except (ImportError, AttributeError):
    pass

from a10_neutron_lbaas.neutron_ext.common import attributes
from a10_neutron_lbaas.neutron_ext.common import constants
from a10_neutron_lbaas.neutron_ext.common import extensions
from a10_neutron_lbaas.neutron_ext.common import resources

RESOURCE_ATTRIBUTE_MAP = resources.apply_template(
    a10_certificate.RESOURCE_ATTRIBUTE_MAP, attributes)

attributes.add_validators(resources.apply_template(
    a10_validators.VALIDATORS, attributes.validators))

LOG = logging.getLogger(__name__)


class A10Certificate(extensions.ExtensionDescriptor):
    nextensions.register_custom_supported_check(
        constants.A10_CERTIFICATE_EXT, lambda: True, plugin_agnostic=True)

    @classmethod
    def get_name(cls):
        return "A10 LBaaS Certificate Management"

    @classmethod
    def get_alias(cls):
        return constants.A10_CERTIFICATE_EXT

    @classmethod
    def get_namespace(cls):
        return "http://docs.openstack.org/ext/neutron/a10-certificates/api/v1.0"

    @classmethod
    def get_updated(cls):
        return "2016-06-19T16:20:13-07:00"

    @classmethod
    def get_description(cls):
        return ("A10 Networks SSL Certificiate Management")

    @classmethod
    def get_resources(cls):
        """Returns external resources."""
        my_plurals = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        attributes.PLURALS.update(my_plurals)
        attr_map = RESOURCE_ATTRIBUTE_MAP
        ext_resources = resource_helper.build_resource_info(my_plurals,
                                                            attr_map,
                                                            constants.A10_CERTIFICATE)

        return ext_resources

    def update_attributes_map(self, attributes):
        super(A10Certificate, self).update_attributes_map(
            attributes,
            extension_attrs_map=RESOURCE_ATTRIBUTE_MAP)

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


@six.add_metaclass(abc.ABCMeta)
class A10CertificatePluginBase(ServicePluginBase):

    def get_plugin_name(self):
        return constants.A10_CERTIFICATE

    def get_plugin_description(self):
        return constants.A10_CERTIFICATE

    def get_plugin_type(self):
        return constants.A10_CERTIFICATE

    def __init__(self):
        super(A10CertificatePluginBase, self).__init__()

    @abc.abstractmethod
    def get_a10_certificates(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def create_a10_certificate(self, context, a10_certificate):
        pass

    @abc.abstractmethod
    def get_a10_certificate(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def update_a10_certificate(self, context, id, certificate):
        pass

    @abc.abstractmethod
    def delete_a10_certificate(self, context, id):
        pass

    @abc.abstractmethod
    def get_a10_certificate_bindings(self, context, filters=None, fields=None):
        pass

    @abc.abstractmethod
    def create_a10_certificate_binding(self, context, a10_certificate_binding):
        pass

    @abc.abstractmethod
    def get_a10_certificate_binding(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def delete_a10_certificate_binding(self, context, id):
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from a10_neutron_lbaas import a10_config
from a10_neutron_lbaas.etc import config as blank_config


def empty_config():
    return a10_config.A10Config(config=blank_config)


def config(config_dict):
    config_constructor = type('config', (object,), config_dict)
    return a10_config.A10Config(config=config_constructor())


def use_config_dir(config_dir=None):
    if config_dir is None:
        config_dir = os.path.dirname(__file__)

    if 'A10_CONFIG_DIR' in os.environ:
        current = os.environ['A10_CONFIG_DIR']

        def cleanup():
            os.environ['A10_CONFIG_DIR'] = current
    else:

        def cleanup():
            del os.environ['A10_CONFIG_DIR']

    os.environ['A10_CONFIG_DIR'] = config_dir

    return cleanup



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid


class FakeModel(object):

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', 'fake-id-001')
        self.tenant_id = kwargs.get('tenant_id', 'get-off-my-lawn')
        self.root_loadbalancer = None
        self.name = kwargs.get("name", "fake-name")


class FakePort(FakeModel):

    def __init__(self, **kwargs):
        self.id = kwargs.get('id', 'fake-port-01')
        self.tenant_id = kwargs.get('tenant_id', "tenantsruinedmypicnic")
        self.name = id


class FakeLoadBalancer(FakeModel):

    def __init__(self, listeners=[]):
        super(FakeLoadBalancer, self).__init__()
        self.id = 'fake-lb-id-001'
        self.listeners = listeners
        self.vip_port_id = str(uuid.uuid4())
        self.vip_port = FakePort()
        self.vip_address = '5.5.5.5'
        self.admin_state_up = True
        self.vip_port = {"id": "vip-id-001", "tenant_id": "tenant_id", "name": "vip-id-001"}
        self.vip_port_id = self.vip_port["id"]
        self.root_loadbalancer = self
        self.vip_subnet_id = "fake-subnet-id-001"
        self.pools = []

    def stats_v21(self):
        self.ret_stats = {
            "bytes_in": 1337,
            "bytes_out": 347,
            "active_connections": 101,
            "total_connections": 1337,
            "extended_stats": {
                "loadbalancer_stat": {
                    "req_bytes": 1337,
                    "resp_bytes": 347,
                    "cur_conns": 101,
                    "tot_conns": 1337,
                    "listener_stat": [{
                        "req_bytes": 1337,
                        "resp_bytes": 347,
                        "name": "3LD3RB33R135",
                        "cur_conns": 101,
                        "tot_conns": 1337,
                        "pool_stat_list": {
                            "cur_conns": 0,
                            "member_stat_list": [{
                                "req_bytes": 0,
                            }]
                        }
                    }]
                }
            }
        }

        self.virt_server = {
            "virtual_server_stat": {
                "req_bytes": 1337,
                "resp_bytes": 347,
                "cur_conns": 101,
                "tot_conns": 1337,
                "vport_stat_list": [{
                    "req_bytes": 1337,
                    "resp_bytes": 347,
                    "cur_conns": 101,
                    "tot_conns": 1337,
                    "name": "3LD3RB33R135"
                }]
            }
        }

        self.virt_service = {
            "virtual_service": {
                "service_group": "3LD3RB33R135"}
        }

        self.serv_group = {
            'service_group_stat': {
                'cur_conns': 0,
                'member_stat_list': [{
                    'req_bytes': 0}]}
        }

    def stats_v30(self):
        self.ret_stats_v30 = {
            "bytes_in": 1337,
            "bytes_out": 347,
            "active_connections": 101,
            "total_connections": 1337,
            "extended_stats": {
                "loadbalancer_stat": {
                    "total_fwd_bytes": 1337,
                    "total_rev_bytes": 347,
                    "curr_conn": 101,
                    "total_conn": 1337,
                    "listener_stat": [{
                        "stats": {
                            "total_fwd_bytes": 1337,
                            "total_rev_bytes": 347,
                            "curr_conn": 101,
                            "total_conn": 1337}}],
                    "pool_stat_list": {
                        'curr_conn': 0,
                        "service_peak_conn": 0,
                        "member_list": [{
                            "stats": {
                                "curr_conn": 0}}]}}}
        }

        self.port_list = {
            "port-list": [{
                "stats": {
                    "total_fwd_bytes": 1337,
                    "total_rev_bytes": 347,
                    "curr_conn": 101,
                    "total_conn": 1337}}]
        }

        self.virt_server = {
            "virtual-server": {
                "port-list": [{
                    "service-group": "3LD3RB33R135"}]}
        }

        self.service_group = {
            "service-group": {
                "stats": {
                    "service_peak_conn": 0}}
        }

        self.members = {
            "member-list": [{
                "stats": {
                    "curr_conn": 0}}]
        }


class FakeListener(FakeModel):

    def __init__(self, protocol, port, admin_state_up=True, pool=None,
                 loadbalancer=None, container_id=None, containers=None):
        super(FakeListener, self).__init__()
        self.id = 'fake-listen-id-001'
        self.name = 'openstackname'
        self.protocol = protocol
        self.protocol_port = port
        self.admin_state_up = admin_state_up
        self.default_pool = pool or FakePool('HTTP', 'ROUND_ROBIN', None)
        self.default_pool_id = self.default_pool.id
        self.loadbalancer = loadbalancer
        self.default_tls_container_id = container_id
        self.sni_containers = containers
        self.root_loadbalancer = FakeLoadBalancer()


class FakePersistence(FakeModel):

    def __init__(self, persistence_type):
        super(FakePersistence, self).__init__()
        self.id = 'fake-pers-id-001'
        self.type = persistence_type


class FakeMember(FakeModel):

    def __init__(self, admin_state_up=True, pool=None,
                 id='fake-member-id-001',
                 address='2.2.2.2'):
        super(FakeMember, self).__init__()
        self.id = id
        self.address = address
        self.admin_state_up = admin_state_up
        self.pool = pool
        self.protocol_port = 80
        self.root_loadbalancer = FakeLoadBalancer()
        self.pool_id = "fake-pool"


class FakePool(FakeModel):

    def __init__(self, protocol, method, persistence, listener=False,
                 members=[], hm=None):
        super(FakePool, self).__init__()
        self.id = 'fake-pool-id-001'
        self.protocol = protocol
        self.lb_algorithm = method
        if persistence is None:
            self.session_persistence = None
        else:
            self.session_persistence = FakePersistence(persistence)
        if listener:
            self.listener = FakeListener(protocol, 2222, pool=self,
                                         loadbalancer=FakeLoadBalancer())
        else:
            self.listener = None
        for member in members:
            member.pool = self
        self.members = members
        self.healthmonitor = hm
        if hm is not None:
            self.healthmonitor.pool = self
        self.root_loadbalancer = FakeLoadBalancer()


class FakeHM(FakeModel):

    def __init__(self, prot, pool=None):
        super(FakeHM, self).__init__()
        self.id = 'fake-hm-id-001'
        self.name = 'hm1'
        self.type = prot
        self.delay = 7
        self.timeout = 7
        self.max_retries = 8
        self.http_method = 'GET'
        self.url_path = '/'
        self.expected_codes = '200'
        self.pool = pool
        self.root_loadbalancer = FakeLoadBalancer()


class FakeCertificateBinding(FakeModel):

    def __init__(self, id=None, certificate_id=None, listener_id=None):
        super(FakeCertificateBinding, self).__init__()
        self.id = id or "binding01"
        self.certificate_id = certificate_id or "certid01"
        self.listener_id = listener_id or 'fake-listen-id-001'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""create orchestration

Revision ID: 2a280aba7701
Revises: 579f359e6e30
Create Date: 2016-04-12 18:25:17.910876

"""

# revision identifiers, used by Alembic.
revision = '2a280aba7701'
down_revision = '579f359e6e30'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'a10_device_instances',
        sa.Column('id', sa.String(36), primary_key=True, nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('tenant_id', sa.String(36), nullable=False),
        sa.Column('name', sa.String(1024), nullable=False),
        sa.Column('username', sa.String(255), nullable=False),
        sa.Column('password', sa.String(255), nullable=False),

        sa.Column('api_version', sa.String(255), nullable=False),
        sa.Column('protocol', sa.String(32), nullable=False),
        sa.Column('port', sa.Integer, nullable=False),
        sa.Column('autosnat', sa.Boolean(), nullable=False),
        sa.Column('v_method', sa.String(32), nullable=False),
        sa.Column('shared_partition', sa.String(1024), nullable=False),
        sa.Column('use_float', sa.Boolean(), nullable=False),
        sa.Column('default_virtual_server_vrid', sa.Integer, nullable=True),
        sa.Column('ipinip', sa.Boolean(), nullable=False),
        sa.Column('write_memory', sa.Boolean(), nullable=False),

        sa.Column('nova_instance_id', sa.String(36), nullable=False),
        sa.Column('host', sa.String(255), nullable=False)
    )
    op.create_table(
        'a10_slbs',
        sa.Column('id', sa.String(36), primary_key=True, nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('tenant_id', sa.String(36), nullable=False),
        sa.Column('device_name', sa.String(1024), nullable=False),
        sa.Column('pool_id', sa.String(36)),
        sa.Column('loadbalancer_id', sa.String(36)),
    )
    pass


def downgrade():
    op.drop_table('a10_slbs')
    op.drop_table('a10_device_instances')
    pass



# Copyright (C) 2014-2015, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import a10_neutron_lbaas
from a10_neutron_lbaas.db.models import a10_certificates as models
from a10_neutron_lbaas.neutron_ext.common import constants
from a10_neutron_lbaas.neutron_ext.db import certificate_db as certs_db
from a10_neutron_lbaas.tests.db import test_base as tbase

import mock
from neutron.plugins.common import constants as nconstants
from neutron.tests.unit.api.v2 import test_base as ntbase

from oslo_log.helpers import logging as logging
from oslo_utils import uuidutils


import os
import os.path


LOG = logging.getLogger(__name__)

_tenant_id = uuidutils.generate_uuid()
_get_path = ntbase._get_path

CERTIFICATE = "a10_certificate"
CERTIFICATES = CERTIFICATE + "s"
CERTIFICATE_BINDING = "a10_certificate_binding"
CERTIFICATE_BINDINGS = CERTIFICATE_BINDING + "s"

CERTIFICATE_EXT = "a10-certificate"


# class CertificateExtensionTestCase(ExtensionTestCase):

#     """Tests a10_openstack.neutron_ext.extensions.Certificates"""

#     def setUp(self):
#         super(CertificateExtensionTestCase, self).setUp()
#         self.plugin = mock.MagicMock()

#         self._setUpExtension(
#             'a10_neutron_lbaas.tests.db.neutron_ext.db.test_certificates.DummyCorePlugin',
#             "", a10Certificate.RESOURCE_ATTRIBUTE_MAP,
#             a10Certificate.A10Certificate, '', supported_extension_aliases='a10-certificate'
#         )

#     def _build_test_binding_collection(self):
#         result = {CERTIFICATE_BINDINGS: []}
#         result[CERTIFICATE_BINDINGS].append(self._build_test_binding(
#             uuidutils.generate_uuid(),
#             uuidutils.generate_uuid(),
#             _tenant_id))
#         result[CERTIFICATE_BINDINGS].append(self._build_test_binding(
#             uuidutils.generate_uuid(),
#             uuidutils.generate_uuid(),
#             _tenant_id))
#         result[CERTIFICATE_BINDINGS].append(self._build_test_binding(
#             uuidutils.generate_uuid(),
#             uuidutils.generate_uuid(),
#             _tenant_id))
#         return result

#     def _build_test_certificate_collection(self):
#         result = {CERTIFICATES: []}

#         result[CERTIFICATES].append(self._build_test_certificate())
#         result[CERTIFICATES].append(self._build_test_certificate(
#             name="Magic Certificate 2"))
#         return result

#     def _build_test_binding(self, certificate_id, vip_id, tenant_id=_tenant_id):
#         result = {'certificate_id': certificate_id,
#                   'vip_id': vip_id}
#         if tenant_id is not None:
#             result['tenant_id'] = tenant_id
#         return result

#     def _build_test_certificate(self, name="Magic Certificate",
#                                 description="Certificate Description",
#                                 cert_data="Super secret certificate data",
#                                 key_data='Key data',
#                                 intermediate_data='Intermediate data',
#                                 tenant_id=_tenant_id,
#                                 password="SecretPassword"):
#         result = {'name': name,
#                   'description': description,
#                   'cert_data': cert_data,
#                   'key_data': key_data,
#                   'intermediate_data': intermediate_data,
#                   'password': password}
#         if tenant_id is not None:
#             result['tenant_id'] = tenant_id
#         return result

#     def test_create_certificate(self):
#         expected = {CERTIFICATE: self._build_test_certificate(_tenant_id)}
#         import pdb
#         pdb.set_trace()
#         # self.plugin.create_a10_certificate = mock.Mock(return_value=expected[CERTIFICATE])

#         self.api.post(_get_path(CERTIFICATES, fmt=self.fmt), self.serialize(expected),
#                       content_type='application/%s' % self.fmt)
#         self.plugin.create_a10_certificate.assert_called_with(mock.ANY, certificate=expected)

#     def test_delete_certificate(self):
#         certificate_id = uuidutils.generate_uuid()
#         self.api.delete(_get_path(CERTIFICATES, id=certificate_id, fmt=self.fmt),
#                         content_type="application/%s" % self.fmt)

#         self.plugin.delete_a10_certificate.assert_called_with(mock.ANY, certificate_id)

#     def test_update_certificate(self):
#         certificate_id = uuidutils.generate_uuid()
#         expected = {CERTIFICATE: self._build_test_certificate(tenant_id=None)}

#         self.plugin.update_a10_certificate = mock.Mock(return_value=expected)
#         self.api.put(_get_path(CERTIFICATES, id=certificate_id, fmt=self.fmt),
#                      self.serialize(expected),
#                      content_type="application/%s" % self.fmt)

#         self.plugin.update_a10_certificate.assert_called_with(mock.ANY,
#                                                               certificate_id,
#                                                               certificate=expected[CERTIFICATE])

#     def test_get_certificate(self):
#         certificate_id = uuidutils.generate_uuid()
#         expected = {CERTIFICATE: self._build_test_certificate()}

#         self.plugin.get_a10_certificate = mock.Mock(return_value=expected[CERTIFICATE])
#         response = self.api.get(_get_path(CERTIFICATES, id=certificate_id, fmt=self.fmt))
#         self.plugin.get_a10_certificate.assert_called_with(
#             mock.ANY, certificate_id, fields=mock.ANY)
#         actual = self.deserialize(response)
#         self.assertEqual(1, len(actual))
#         self.assertEqual(expected, actual)
#         self.assertEqual(expected[CERTIFICATE]['name'], actual[CERTIFICATE]['name'])

#     def test_get_certificates(self):
#         expected = self._build_test_certificate_collection()

#         self.plugin.get_a10_certificates = mock.Mock(return_value=expected[CERTIFICATES])
#         url = _get_path(CERTIFICATES, fmt=self.fmt)
#         response = self.api.get(url)

#         self.plugin.get_a10_certificates.assert_called_with(mock.ANY,
#                                                             fields=mock.ANY,
#                                                             filters=mock.ANY)
#         actual = self.deserialize(response)
#         self.assertEqual(expected[CERTIFICATES], actual[CERTIFICATES])

#     def test_get_certificate_binding(self):
#         binding_id = uuidutils.generate_uuid()
#         certificate_id = uuidutils.generate_uuid()
#         vip_id = uuidutils.generate_uuid()
#         expected = {CERTIFICATE_BINDING: self._build_test_binding(certificate_id, vip_id)}

#         self.plugin.get_a10_certificate_binding = mock.Mock(
#             return_value=expected[CERTIFICATE_BINDING])
#         url = _get_path(CERTIFICATE_BINDINGS, id=binding_id, fmt=self.fmt)
#         response = self.api.get(url)
#         self.plugin.get_a10_certificate_binding.assert_called_with(
#             mock.ANY, binding_id, fields=mock.ANY)
#         actual = self.deserialize(response)
#         self.assertEqual(expected, actual)
#         self.assertEqual("200 OK", response._status)

#     def test_get_certificate_bindings(self):
#         expected = self._build_test_binding_collection()

#         self.plugin.get_a10_certificate_bindings = mock.Mock(
#             return_value=expected[CERTIFICATE_BINDING])
#         url = _get_path(CERTIFICATE_BINDINGS, fmt=self.fmt)
#         result = self.api.get(url)
#         actual = self.deserialize(result)

#         self.plugin.get_a10_certificate_bindings.assert_called_with(mock.ANY,
#                                                                     fields=mock.ANY,
#                                                                     filters=mock.ANY)
#         self.assertEqual(expected, actual)

#     def test_create_certificate_binding(self):
#         certificate_id = uuidutils.generate_uuid()
#         vip_id = uuidutils.generate_uuid()
#         expected = {CERTIFICATE_BINDING: self._build_test_binding(certificate_id, vip_id)}

#         self.plugin.create_a10_certificate_binding = mock.Mock(return_value=expected)
#         url = _get_path(CERTIFICATE_BINDINGS, fmt=self.fmt)
#         self.api.post(url, self.serialize(expected),
#                       content_type='application/%s' % self.fmt)
#         self.plugin.create_a10_certificate_binding.assert_called_with(mock.ANY,
#                                                                       certificate_binding=expected)

#     def test_delete_certificate_binding(self):
#         binding_id = uuidutils.generate_uuid()
#         url = _get_path(CERTIFICATE_BINDINGS, id=binding_id, fmt=self.fmt)
#         self.api.delete(url, content_type="application/%s" % self.fmt)
#         self.plugin.delete_a10_certificate_binding.assert_called_with(mock.ANY, binding_id)


class CertificateDbMixInTestCase(tbase.UnitTestBase):

    """Tests a10_openstack.neutron_ext.db.certificate_db.CertificateManager"""
    _test_name = 'THE CERTIFICATE'
    _test_description = 'THE CERTIFICATE DESCRIBED'
    _test_cert_type = 'Certificate'
    _test_cert_data = 'THE CERT DATA THAT SHOULD BE ENCRYPTED'
    _root_dir = os.path.dirname(os.path.dirname(__file__))
    _etc_dir = os.path.join(_root_dir, 'etc')
    # TODO(mdurrant) Make fake certs and read them in.
    _test_key_data = 'fake key'
    _test_intermediate_data = 'fake intermediate'
    _test_password = 'password'

    # figure out where we're being loaded from
    _base_path = os.path.abspath(a10_neutron_lbaas.__file__)
    _test_path = _base_path[:_base_path.rfind('/')]
    _test_etc_path = os.path.abspath("{0}/tests/etc".format(_test_path))
    _test_conf_path = "{0}/{1}".format(_test_etc_path, 'neutron.conf.test')

    # _test_conf_path =

    def setUp(self):
        super(CertificateDbMixInTestCase, self).setUp()
        self._nm_patcher = mock.patch('neutron.manager.NeutronManager')
        nm = self._nm_patcher.start()
        # nm.get_service_plugins.return_value = {
        #     nconstants.LOADBALANCERV2: mock.MagicMock()
        # }

        self.plugin = certs_db.A10CertificateDbMixin()

    def tearDown(self):
        super(CertificateDbMixInTestCase, self).tearDown()

    def context(self):
        session = self.open_session()
        context = mock.Mock(session=session, tenant_id='fake-tenant-id')
        return context

    def _build_certificate(self, name=_test_name, description=_test_description,
                           cert_data=_test_cert_data, key_data=_test_key_data,
                           intermediate_data=_test_intermediate_data,
                           tenant_id='tenant1', password=_test_password):
        return {'a10_certificate': {'name': name,
                                    'description': description,
                                    'cert_data': cert_data,
                                    'key_data': key_data,
                                    'intermediate_data': intermediate_data,
                                    'password': password,
                                    'tenant_id': tenant_id}}

    def _build_binding(self, certificate_id=None, listener_id=None, tenant_id=None):
        return {'a10_certificate_binding': {'certificate_id': certificate_id,
                                            'listener_id': listener_id,
                                            'tenant_id': tenant_id}}

    def _create_certificate(self, tenant_id='tenant1', ctx=None, description=None):
        if ctx is None:
            ctx = self.context()
        ctx.tenant_id = tenant_id
        certificate = self._build_certificate()
        return self.plugin.create_a10_certificate(ctx, certificate), certificate

    def _create_binding(self, ctx=None, cert_id=None, listener_id=None, tenant_id=_tenant_id):
        if ctx is None:
            ctx = self.context()
        ctx.tenant_id = tenant_id
        binding = self._build_binding(ctx, cert_id, listener_id, tenant_id)

        return self.plugin.create_a10_certificate_listener_binding(ctx, binding), binding

    def test_create_certificate(self, tenant_id="tenant1"):
        ctx = self.context()
        record, request = self._create_certificate(tenant_id)
        res = ctx.session.query(models.Certificate).all()
        self.assertEqual(1, len(res))
        actual = res[0]
        self.assertFalse(actual is None)
        self.assertEqual(self._test_name, actual.name)
        self.assertEqual(self._test_description, actual.description)
        self.assertEqual(self._test_cert_data, actual.cert_data)
        self.assertEqual(self._test_key_data, actual.key_data)
        self.assertEqual(self._test_intermediate_data, actual.intermediate_data)
        self.addCleanup(ctx.session.delete, actual)

    def test_update_certificate(self):
        ctx = self.context()
        cert, certificate = self._create_certificate(ctx=ctx)
        certificate = {'a10_certificate': {'name': 'THE SECOND CERTIFICATE',
                                           'description': 'THE SECOND CERT DESCRIBED',
                                           'cert_data': 'M0re s00p3r s3cr3t d@ta',
                                           'key_data': 'Key data',
                                           'intermediate_data': 'intermediate data',
                                           'password': 'password'}}
        self.plugin.update_a10_certificate(ctx, cert['id'], certificate)
        actual = (ctx.session.query(models.Certificate).filter_by(id=cert['id']).one())
        self.assertEqual('THE SECOND CERTIFICATE', actual.name)
        self.assertEqual('THE SECOND CERT DESCRIBED', actual.description)
        self.assertEqual('M0re s00p3r s3cr3t d@ta', actual.cert_data)

    def test_delete_certificate(self):
        ctx = self.context()
        cert, data = self._create_certificate(ctx=ctx)
        self.plugin.delete_a10_certificate(ctx, cert['id'])
        actual = (ctx.session.query(models.Certificate).filter_by(id=cert['id']).first())
        self.assertTrue(actual is None)

    def test_get_certificate(self):
        ctx = self.context()
        cert, data = self._create_certificate()
        actual = self.plugin.get_a10_certificate(ctx, cert['id'])
        self.assertEqual(cert, actual)

    def test_get_throwsexceptionwhenitdoesntexist(self):
        ctx = self.context()
        self.assertRaises(certs_db.CertificateNotFoundError, self.plugin.get_a10_certificate,
                          (self, ctx, uuidutils.generate_uuid()), None)

    def test_create_certificate_binding(self):
        ctx = self.context()
        vip_uuid = 'fake-listener-id'
        cert, data = self._create_certificate()
        binding = self._build_binding(cert['id'], vip_uuid)
        self.plugin.create_a10_certificate_binding(ctx, binding)

        actual = ctx.session.query(models.CertificateListenerBinding).filter_by(
            listener_id=vip_uuid, certificate_id=cert['id']).first()

        self.assertEqual(cert['id'], actual.certificate_id)
        self.assertEqual(vip_uuid, actual.listener_id)

    def test_delete_certificate_binding(self):
        ctx = self.context()
        vip_uuid = 'fake-listener-id'
        cert, data = self._create_certificate()
        binding = self._build_binding(cert['id'], vip_uuid)
        self.plugin.create_a10_certificate_binding(ctx, binding)
        ctx.session.commit()
        record = ctx.session.query(models.CertificateListenerBinding) \
            .filter_by(listener_id=vip_uuid, certificate_id=cert['id']).first()

        self.plugin.delete_a10_certificate_binding(ctx, record.id)
        actual = ctx.session.query(models.CertificateListenerBinding).filter_by(
            listener_id=vip_uuid, certificate_id=cert['id']).first()
        self.assertTrue(actual is None)


# TODO(mdurrant) - Move this in to tests or tests/unit
class DummyCorePlugin(object):

    """Dummy Neutron Core plugin for unit testing"""
    pass


# TODO(mdurrant) - Move this in to tests or tests/unit
class DummyServicePlugin(object):

    """Dummy Neutron Service plugin for unit testing"""

    def driver_loaded(self, driver):
        pass

    def get_plugin_type(self):
        return constants.DUMMY

    def get_plugin_description(self):
        return "Dummy service plugin"


# TODO(mdurrant) - Move this in to tests or tests/unit
class DummyServiceDriver(object):

    """Dummy Neutron Service driver for unit testing"""
    @staticmethod
    def get_service_type():
        return constants.DUMMY

    def __init__(self, plugin):
        pass



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

verify_appliances = True

devices = {
    "ax1": {
        "name": "ax1",
        "host": "10.10.100.20",
        "port": 8443,
        "protocol": "https",
        "username": "admin",
        "password": "a10",
        "status": True,
        "autosnat": True,
        "api_version": "2.1",
        "v_method": "LSI",
        "max_instance": 5000,
        "use_float": True,
        "method": "hash"
    },
    "ax2": {
        "name": "ax2",
        "host": "10.10.100.21",
        "port": 8080,
        "protocol": "http",
        "username": "admin",
        "password": "a10",
        "status": False,
        "api_version": "2.1",
        "v_method": "LSI",
        "max_instance": 5000,
        "use_float": True,
        "method": "hash"
    },
    "ax3": {
        "host": "10.10.100.22",
        "protocol": "http",
        "username": "admin",
        "password": "a10",
        "status": True,
        "api_version": "2.1",
        "max_instance": 5000,
        "use_float": True,
    },
    "ax4": {
        "host": "10.10.100.23",
        "username": "admin",
        "password": "a10",
        "api_version": "2.1",
        "use_float": True,
        "ha_sync_list": [
            {
                "name": "ax5",
                "ip": "1.1.1.1",
                "username": "admin",
                "password": "a10"
            }
        ]
    },
    "axxv21": {
        "host": "10.10.100.29",
        "protocol": "http",
        "username": "admin",
        "password": "a10",
        "status": True,
        "api_version": "2.1",
        "max_instance": 5000,
        "use_float": True,
    },
    "axv30": {
        "host": "10.10.100.30",
        "protocol": "http",
        "username": "admin",
        "password": "a10",
        "status": True,
        "api_version": "3.0",
        "max_instance": 5000,
        "use_float": True,
    },
    # "axadp": {
    #     "host": "10.10.100.24",
    #     "username": "admin",
    #     "password": "a10",
    #     "protocol": "https",
    #     "v_method": "ADP",
    # },
    # "axadp-badalt": {
    #     "host": "10.10.100.24",
    #     "username": "admin",
    #     "password": "a10",
    #     "protocol": "https",
    #     "v_method": "ADP",
    #     "shared_partition": "mypart",
    # },
    "axadp-alt": {
        "host": "10.10.100.24",
        "username": "admin",
        "password": "a10",
        "protocol": "https",
        "v_method": "LSI",
        "shared_partition": "mypart",
    },
    "axadp-noalt": {
        "host": "10.10.100.24",
        "username": "admin",
        "password": "a10",
        "protocol": "https",
        "v_method": "LSI"
    },
    "ax-nowrite": {
        "host": "10.10.100.24",
        "username": "admin",
        "password": "a10",
        "protocol": "https",
        "write_memory": False,
    },
    "ax-write": {
        "host": "10.10.100.24",
        "username": "admin",
        "password": "a10",
        "protocol": "https",
    },
    "axipinip": {
        "host": "10.48.5.219",
        "protocol": "https",
        "username": "admin",
        "password": "a10",
        "status": True,
        "api_version": "2.1",
        "max_instance": 5000,
        "use_float": True,
        "ipinip": True
    },
}

vthunder = {
    'username': 'admin',
    'password': 'a10',

    'api_version': '3.0',

    'nova_flavor': 'acos.min',
    'glance_image': '80bf6b06-8481-485b-a3ae-87a21bde2438',

    'vthunder_tenant_name': 'admin',
    'vthunder_tenant_username': 'admin',
    'vthunder_tenant_password': 'password',

    'vthunder_management_network': 'private',
    'vthunder_data_networks': ['vipnet', 'membernet'],

    'sflow_collector': {"host": "10.10.10.10", "port": 6343},

    'license_manager': {
        "hosts": [
                {"ip": "licenseserver.com", "port": 443}
        ],
        "serial": "SNTHISISNOTREAL",
        "instance-name": "UNITTESTINSTANCE",
        "bandwidth-base": 100,
        "interval": 3,
        "use-mgmt-port": True
    },
    'write_memory': False

}

image_defaults = {
    "name": None,
    "id": None,
    "visibility": "private",
    "tags": ["a10"],
    "properties": None,
    "container_format": "bare",
    "disk_format": "qcow2",
    "min_disk": 10,
    "min_ram": 4096,
    "protected": False
}

instance_defaults = {
    "flavor": "acos.min",
    "networks": ["private", "private", "private"]
}

neutron_config = "/etc/neutron/neutron.conf"
scaling_monitor = {"ip": "127.0.0.1", "port": 2358}

# For provider based config testing

who_should_win = 'daleks'
best_spaceship = 'tardis'

providers = {
    'prov1': {
        'who_should_win': 'the-doctor',
        'vthunder': {
            'api_version': '9.9'
        }
    }
}


# Name-based dynamic configuration
vport_expressions = {
    "descriptor": {
        "regex": "^ABC123",
        "json": {
            "setting1": True
        }
    }
}

virtual_server_expressions = {
    "descriptor": {
        "regex": "^ABC123",
        "json": {
            "setting1": True
        }
    }
}

service_group_expressions = {
    "descriptor": {
        "regex": "^ABC123",
        "json": {
            "setting1": True
        }
    }
}

member_expressions = {
    "descriptor": {
        "regex": "^ABC123",
        "json": {
            "setting1": True
        }
    }
}

monitor_expressions = {
    "descriptor": {
        "regex": "^ABC123",
        "json": {
            "setting1": True
        }
    }
}



# Copyright (C) 2016, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""v2 SSL Certificates

Revision ID: 08dacae1fb67
Revises: 3c7123f2aeba
Create Date: 2016-06-21 06:18:58.773598

"""

# revision identifiers, used by Alembic.
revision = '08dacae1fb67'
down_revision = '3c7123f2aeba'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'a10_certificates',
        sa.Column('tenant_id', sa.String(255), nullable=True),
        sa.Column('id', sa.String(36), primary_key=True, nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('name', sa.String(255), nullable=False),
        sa.Column('description', sa.Text(1024), nullable=True),
        sa.Column('cert_data', sa.Text(8000), nullable=False),
        sa.Column('key_data', sa.Text(8000), nullable=False),
        sa.Column('intermediate_data', sa.Text(8000), nullable=True),
        sa.Column('password', sa.String(1024), nullable=True),
    )

    op.create_table(
        'a10_certificatelistenerbindings',
        sa.Column('id', sa.String(36), primary_key=True, nullable=False),
        sa.Column('created_at', sa.DateTime, nullable=False),
        sa.Column('updated_at', sa.DateTime, nullable=False),
        sa.Column('tenant_id', sa.String(36), nullable=False),
        sa.Column('certificate_id', sa.String(36),
                  sa.ForeignKey(u'a10_certificates.id'),
                  nullable=False),
        sa.Column('listener_id', sa.String(36), nullable=False)
    )


def downgrade():
    op.drop_table('a10_certificatelistenerbindings')
    op.drop_table('a10_certificates')



# Copyright 2016,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections

import attributes

import six


def remove_attributes_not_specified(resource):
    return dict((k, v) for k, v in resource.items() if v != attributes.ATTR_NOT_SPECIFIED)


def apply_template(template, *args, **kw):
    """Applies every callable in any Mapping or Iterable"""
    if six.callable(template):
        return template(*args, **kw)
    if isinstance(template, six.string_types):
        return template
    if isinstance(template, collections.Mapping):
        return template.__class__((k, apply_template(v, *args, **kw)) for k, v in template.items())
    if isinstance(template, collections.Iterable):
        return template.__class__(apply_template(v, *args, **kw) for v in template)
    return template



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import sys

import a10_neutron_lbaas.tests.unit.mocks as mocks

import a10_neutron_lbaas.a10_exceptions as a10_ex
import a10_neutron_lbaas.tests.unit.test_base as test_base

# Figure out why this is necessary to make the mock work correctly.

log_mock = mock.MagicMock()

# This is here because I couldn't get mock to actually mock the classes
# Get rid of this and I'll give you a cookie - MD
sys.modules["a10_neutronclient"] = mock.MagicMock()
sys.modules["a10_openstack_lib"] = mock.MagicMock()
sys.modules["glanceclient"] = mock.MagicMock()
sys.modules["glanceclient.client"] = mock.MagicMock()
sys.modules["logging"] = mock.Mock(getLogger=mock.Mock(return_value=log_mock))
sys.modules["django"] = mock.MagicMock()
sys.modules["django.conf"] = mock.MagicMock()
sys.modules["django.conf.settings"] = mock.MagicMock()
sys.modules["openstack_dashboard"] = mock.MagicMock()
sys.modules["openstack_dashboard.api"] = mock.MagicMock()
sys.modules["openstack_dashboard.api.neutron"] = mock.MagicMock()
sys.modules["openstack_dashboard.api.neutron.NeutronAPIDictWrapper"] = mock.MagicMock()


# a10 client that extends neutronclient.v2_0.client.Client

import a10_neutron_lbaas.vthunder.instance_manager as im

VMSTATE_KEY = "OS-EXT-STS:vm_state"
DEFAULT_INSTANCE_STATES = [
    "INITALIZED",
    "ACTIVE"
]

instance_states = []


class TestInstanceManager(test_base.UnitTestBase):
    def setUp(self):
        log_mock.reset_mock()
        super(TestInstanceManager, self).setUp()

        im.CREATE_TIMEOUT = 10

        glance_patch = mock.patch("glanceclient.client")
        self.glance_api = glance_patch.start()
        self.addCleanup(glance_patch.stop)

        keystone_patch = mock.patch("keystoneclient.client.Client")
        self.keystone_api = keystone_patch.start()
        self.addCleanup(keystone_patch.stop)

        neutron_patch = mock.patch("neutronclient.neutron.client.Client")
        self.neutron_api = neutron_patch.start()
        self.addCleanup(neutron_patch.stop)

        nova_patch = mock.patch("novaclient.client.Client")
        self.nova_api = nova_patch.start()
        self.addCleanup(nova_patch.stop)
        self.instance_states = ["ACTIVE",
                                "ACTIVE"]

        self.setup_mocks()
        self.ks_version = 3
        self.auth_url = "http://localhost:1234"
        self.vthunder_tenant_name = "tenantive-tenant"
        self.user = "user"
        self.password = "password"
        self.target = im.InstanceManager(self.ks_version, self.auth_url, self.vthunder_tenant_name,
                                         self.user, self.password, nova_api=self.nova_api,
                                         neutron_api=self.neutron_api)

    def server_get_side_effect(self, args):
        if len(instance_states) > 0:
            vmstate = instance_states.pop()
        else:
            vmstate = "ACTIVE"
        setattr(self.fake_created, VMSTATE_KEY, vmstate)

    def setup_mocks(self):
        self.fake_created = mock.Mock(addresses={"mgmt-net": [
            {
                "version": 4,
                "addr": "127.0.0.1"
            }
        ]})
        setattr(self.fake_created, VMSTATE_KEY, "ACTIVE")

        self.a10_context = mock.MagicMock()
        self.nova_api.servers = mock.Mock()
        self.fake_image = self._fake_image()
        self.fake_flavor = self._fake_flavor()
        self.fake_networks = {
            "networks": [
                {"id": "mgmt-net", "name": "mgmt-net"}]
        }

        self.fake_instance = self._fake_instance(name="fake-instance-01",
                                                 image=self.fake_image.id,
                                                 flavor=self.fake_flavor.id,
                                                 nics=['mgmt-net'])

        self.nova_api.servers.list = mock.Mock(return_value=[self._fake_instance()])
        self.nova_api.servers.create = mock.Mock(return_value=self.fake_created)
        self.nova_api.servers.get = mock.Mock(return_value=self.fake_created)

        self.nova_api.flavors = mock.Mock()
        self.nova_api.flavors.list = mock.Mock(return_value=[self._fake_flavor()])

        self.nova_api.images = mock.Mock()
        self.nova_api.images.list = mock.Mock(return_value=[self._fake_image()])

        self.glance_api.images.list.return_value = [self._fake_image()]

        self.neutron_api.list_networks = mock.Mock(return_value=self.fake_networks)

        self.neutron_api.get_a10_instances = mock.Mock(return_value=[{}])
        self.neutron_api.create_a10_instance = mock.Mock(return_value=self.fake_created)

    def _fake_instance(self, id=1, name="instance001", image=None, flavor=None, nics=[]):
        return mocks.FakeInstance(id=id, name=name, image=image, flavor=flavor, nics=nics)

    def _fake_flavor(self, id=1, name="flavor001"):
        return mocks.FakeFlavor(id=id, name=name)

    def _fake_image(self, id=1, name="image001"):
        import json

        metadata = {
            "properties": json.dumps(
                {
                    "api_version": 2.1,
                    "username": "dude",
                    "password": "password",
                    "protocol": "http",
                    "port": 80
                }
            )
        }
        return mocks.FakeImage(id=id, name=name, metadata=metadata)

    def _test_create(self, instance):
        self.target.create_instance(instance)

    def test_create_calls_nova_api(self):
        self._test_create(self.fake_instance)
        self.assertTrue(self.nova_api.servers.create.called)

    def test_create_calls_nova_api_with_args(self):
        self._test_create(self.fake_instance)
        expected = self.fake_instance.copy()
        expected["nics"] = [{'net-id': x} for x in self.fake_instance["networks"]]
        del expected["networks"]

        self.nova_api.servers.create.assert_called_with(**expected)

    def test_get_instance_gets_image_get(self):
        instance_id = "INSTANCE_ID"
        self.nova_api.servers.get.side_effect = self.server_get_side_effect
        self.target.get_instance(instance_id)
        self.nova_api.flavors.list.assert_called()

    def test_delete_instance_calls_nova(self):
        fake_instance = self._fake_instance()
        self.target.delete_instance(fake_instance["id"])
        self.nova_api.servers.delete.assert_called_with(fake_instance["id"])

    def test_list_instances_calls_nova_list(self):
        detailed = True
        search_opts = None
        marker = None
        limit = None
        sort_keys = None
        sort_dirs = None

        self.target.list_instances(detailed, search_opts, marker, limit, sort_keys, sort_dirs)
        self.nova_api.servers.list.assert_called_with(detailed, search_opts, marker, limit,
                                                      sort_keys, sort_dirs)

    def test_neutronapi_notnone(self):
        self.assertIsNotNone(self.target._neutron_api)

    def test_novaapi_notnone(self):
        self.assertIsNotNone(self.target._nova_api)

    def test_create_instance_flavor_not_available_throws_exception(self):
        self.nova_api.flavors.list.return_value = [None]
        fake_instance = self.fake_instance
        with self.assertRaises(a10_ex.FlavorNotFoundError):
            self.target.create_instance(fake_instance)

    def test_create_instance_network_not_available_throws_exception(self):
        self.neutron_api.list_networks.return_value = {"networks": [None]}
        self.neutron_api.list_networks.side_effect = a10_ex.NetworksNotFoundError()
        fake_instance = self.fake_instance
        fake_instance["networks"] = ["fakenet"]

        with self.assertRaises(a10_ex.NetworksNotFoundError):
            self.target.create_instance(fake_instance)

    def test_get_network_throws_exception_for_unspecified_identifier(self):
        with self.assertRaises(a10_ex.IdentifierUnspecifiedError):
            self.target.get_networks(networks={})

    def test_get_image_throws_exception_for_unspecified_identifier(self):
        with self.assertRaises(a10_ex.IdentifierUnspecifiedError):
            self.target.get_image(identifier=None)

    def test_get_flavor_throws_exception_for_unspecified_identifier(self):
        with self.assertRaises(a10_ex.IdentifierUnspecifiedError):
            self.target.get_flavor(identifier=None)

    def test_get_flavor_throws_exception_for_missing_flavor(self):
        self.nova_api.flavors.list.return_value = [None]
        self.nova_api.flavors.list.side_effect = a10_ex.FlavorNotFoundError
        with self.assertRaises(a10_ex.FlavorNotFoundError):
            self.target.get_flavor(identifier="invalidflavor")

    def test_get_networks_failure_returns_exception(self):
        self.neutron_api.list_networks.side_effect = a10_ex.ServiceUnavailableError()
        self.neutron_api.list_networks.return_value = {"networks": []}

        with self.assertRaises(a10_ex.NetworksNotFoundError):
            self.target.get_networks(networks=["network"])

    def test_create_device_instance_returns_host(self):
        defaults = im._default_server
        defaults["flavor"] = "flavor001"
        defaults["networks"] = ["mgmt-net"]
        defaults["glance_image"] = "image001"
        defaults["nova_flavor"] = "flavor001"
        defaults["vthunder_management_network"] = "mgmt-net"
        defaults["vthunder_data_networks"] = ["mgmt-net"]

        self.fake_created.addresses = dict((network, [
            {
                "version": 4,
                "addr": "127.0.0.1"
            }
        ]) for network in defaults['networks'])

        self.nova_api.flavors.list.return_value = [self._fake_flavor(
            name=defaults['flavor'])]

        self.neutron_api.list_networks.return_value = {
            "networks": [{"id": x, "name": x}
                         for x in defaults['networks']]
        }

        device = self.target.create_device_instance(defaults)
        self.assertEqual('127.0.0.1', device['ip_address'])



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import acos_client.errors as acos_errors
import logging

from a10_neutron_lbaas.acos import openstack_mappings
import handler_base_v1
import v1_context as a10


LOG = logging.getLogger(__name__)


class HealthMonitorHandler(handler_base_v1.HandlerBaseV1):

    def _name(self, hm):
        return hm['id'][0:28]

    def _set(self, c, set_method, context, hm):
        hm_name = self._meta_name(hm)
        method = None
        url = None
        expect_code = None
        if hm['type'] in ['HTTP', 'HTTPS']:
            method = hm['http_method']
            url = hm['url_path']
            expect_code = hm['expected_codes']

        args = self.meta(hm, 'hm', {})

        set_method(hm_name, openstack_mappings.hm_type(c, hm['type']),
                   hm['delay'], hm['timeout'], hm['max_retries'],
                   method=method, url=url, expect_code=expect_code,
                   axapi_args=args)

    def create(self, context, hm, pool_id):
        h = hm.copy()
        h['pool_id'] = pool_id
        with a10.A10WriteHMStatusContext(self, context, h, action='create') as c:
            try:
                self._set(c, c.client.slb.hm.create, context, hm)
            except acos_errors.Exists:
                pass

            if pool_id is not None:
                c.client.slb.service_group.update(
                    self._pool_name(context, pool_id),
                    health_monitor=self._meta_name(hm))

            for pool in hm['pools']:
                if pool['pool_id'] == pool_id:
                    continue
                c.client.slb.service_group.update(
                    self._pool_name(context, pool['pool_id']),
                    health_monitor=self._meta_name(hm))

    def update(self, context, old_hm, hm, pool_id):
        h = hm.copy()
        h['pool_id'] = pool_id
        with a10.A10WriteHMStatusContext(self, context, h) as c:
            self._set(c, c.client.slb.hm.update, context, hm)

    def _dissociate(self, c, context, hm, pool_id):
        """Remove a pool association"""
        pool_name = self._pool_name(context, pool_id)
        c.client.slb.service_group.update(pool_name, health_monitor="",
                                          health_check_disable=True)

    def dissociate(self, c, context, hm, pool_id):
        """Remove a pool association, and the healthmonitor if its the last one"""

        self._dissociate(c, context, hm, pool_id)
        pools = hm.get("pools", [])
        if not any(p for p in pools if p.get("pool_id") != pool_id):
            self._delete_unused(c, context, hm)

    def _delete(self, c, context, hm):
        """Delete a healthmonitor and ALL its pool associations"""
        pools = hm.get("pools", [])

        for pool in pools:
            pool_id = pool.get("pool_id")
            self._dissociate(c, context, hm, pool_id)

        self._delete_unused(c, context, hm)

    def _delete_unused(self, c, context, hm):
        try:
            c.client.slb.hm.delete(self._meta_name(hm))
        except acos_errors.InUse:
            LOG.error("Cannot delete a health monitor with existing associations")
            raise
        except acos_errors.NotFound:
            pass

    def delete(self, context, hm, pool_id):
        h = hm.copy()
        # Get the binding count to see if we need to perform disassociation

        h['pool_id'] = pool_id
        with a10.A10DeleteHMContext(self, context, h) as c:

            if pool_id is None:
                # Delete the whole healthmonitor
                self._delete(c, context, hm)
            else:
                # Disassociate this pool
                self.dissociate(c, context, hm, pool_id)



# Copyright (C) 2016, A10 Networks Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Provides backwards compatibility for neutron -> neutron_lib moves"""

try:
    # Get the ExtensionDescriptor class from neutron
    from neutron.api.extensions import ExtensionDescriptor as ExtensionDescriptor  # noqa
except (AttributeError, ImportError):
    # If it's not there, try neutron_lib. Else, we're hosed.
    # Only way we can deal with such breaking changes
    from neutron_lib.api.extensions import ExtensionDescriptor as ExtensionDescriptor  # noqa



# Copyright 2015-2016, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.from neutron.db import model_base

import mock
import uuid

import a10_openstack_lib.resources.a10_device_instance as resources

import a10_neutron_lbaas.a10_config as a10_config
import a10_neutron_lbaas.neutron_ext.common.constants as constants
from a10_neutron_lbaas.neutron_ext.extensions import a10DeviceInstance
import a10_neutron_lbaas.neutron_ext.services.a10_device_instance.plugin as plugin
from a10_neutron_lbaas.tests.db.neutron_ext.db import test_a10_device_instance


class TestPlugin(test_a10_device_instance.TestA10DeviceInstanceDbMixin):

    def setUp(self):
        super(TestPlugin, self).setUp()
        self.plugin = plugin.A10DeviceInstancePlugin()
        self.target = self.plugin

        self.maxDiff = None

        self.instance_manager = mock.MagicMock()
        self.instance_manager.create_device_instance.side_effect = self.fake_instance

        self.vthunder_defaults = a10_config.A10Config().get_vthunder_config()

        self._im_patcher = mock.patch(
            'a10_neutron_lbaas.vthunder.instance_manager.InstanceManager.from_config',
            return_value=self.instance_manager)
        self._im_patcher.start()

    def tearDown(self):
        self._im_patcher.stop()
        super(TestPlugin, self).tearDown()

    def fake_instance(self, *args, **kwargs):
        return {
            'nova_instance_id': str(uuid.uuid4()),
            'ip_address': '10.11.12.13'
        }

    def fake_deviceinstance(self):
        return {
            'name': 'fake-name',
            'host': 'fake-host',
            'api_version': 'fake-version',
            'username': 'fake-username',
            'password': 'fake-password',
            'protocol': 'http',
            'port': 12345
        }

    def blank(self, resource):
        attr_map = resources.RESOURCE_ATTRIBUTE_MAP[resource]
        return dict((k, attr_map[k]['default'])
                    for k in attr_map
                    if 'default' in attr_map[k]
                    and not callable(attr_map[k]['default'])
                    and attr_map[k]['is_visible'])

    def default_options(self):
        return self.blank(resources.RESOURCES)

    def test_create_a10_device_instance(self):
        instance = {}
        context = self.context()
        result = self.plugin.create_a10_device_instance(context, self.envelope(instance))
        self.assertIsNotNone(result['id'])

        expected = self.default_options()
        expected.update(instance)
        expected.update(
            {
                'id': result['id'],
                'nova_instance_id': result['nova_instance_id'],
                'host': result['host'],
                'tenant_id': context.tenant_id,
                'project_id': context.tenant_id,
                'api_version': self.vthunder_defaults['api_version'],
                'username': self.vthunder_defaults['username'],
                'password': self.vthunder_defaults['password'],
                'protocol': self.vthunder_defaults['protocol'],
                'port': self.vthunder_defaults['port'],
                'description': ''
            })

        self.assertEqual(expected, result)

    def test_create_a10_device_instance_options(self):
        instance = self.fake_deviceinstance()
        context = self.context()
        result = self.plugin.create_a10_device_instance(context, self.envelope(instance))
        self.assertIsNotNone(result['id'])

        expected = self.default_options()
        expected.update(instance)
        expected.update(
            {
                'id': result['id'],
                'nova_instance_id': result['nova_instance_id'],
                'host': result['host'],
                'tenant_id': context.tenant_id,
                'project_id': context.tenant_id
            })

        self.assertEqual(expected, result)

    def test_update_a10_device_instance_options(self):
        instance = self.default_options()
        create_context = self.context()
        create_result = self.plugin.create_a10_device_instance(create_context,
                                                               self.envelope(instance))
        self.assertIsNotNone(create_result['id'])

        request = self.fake_deviceinstance()
        context = self.context()
        result = self.plugin.update_a10_device_instance(context,
                                                        create_result['id'],
                                                        self.envelope(request))

        expected = create_result.copy()
        expected.update(request)

        self.assertEqual(expected, result)

    def test_get_a10_device_instance(self):
        instance = self.fake_deviceinstance()
        create_context = self.context()
        create_result = self.plugin.create_a10_device_instance(create_context,
                                                               self.envelope(instance))

        context = self.context()
        result = self.plugin.get_a10_device_instance(context, create_result['id'])

        self.assertEqual(create_result, result)

    def test_get_a10_device_instance_not_found(self):
        context = self.context()
        self.assertRaises(
            a10DeviceInstance.A10DeviceInstanceNotFoundError,
            self.plugin.get_a10_device_instance,
            context,
            'fake-deviceinstance-id')

    def test_get_a10_device_instances(self):
        instance = self.fake_deviceinstance()
        create_context = self.context()
        create_result = self.plugin.create_a10_device_instance(create_context,
                                                               self.envelope(instance))
        create_context.session.commit()

        context = self.context()
        result = self.plugin.get_a10_device_instances(context)

        self.assertEqual([create_result], result)

    def _build_instance(self):
        return {
            "a10_device_instance": {
                "name": "asdf",
                "host": "10.10.42.42",
                "image": "MY_FAKE_IMAGE",
                "flavor": "MY_FAKE_FLAVOR",
                "management_network": "this_network",
                "data_networks": ["that_network"],
            }
        }

    def test_supported_extension_aliases(self):
        sea = self.plugin.supported_extension_aliases
        self.assertEqual([constants.A10_DEVICE_INSTANCE_EXT], sea)

    def test_create_calls_instance_manager(self):
        context = self.context()
        context.session = mock.MagicMock()
        self.target.create_a10_device_instance(context, self._build_instance())
        self.assertTrue(self.instance_manager.create_device_instance.called)

    def test_delete_calls_instance_manager(self):
        context = self.context()
        context.session = mock.MagicMock()
        self.target.delete_a10_device_instance(context, 1)
        delete_call = self.instance_manager.delete_instance
        self.assertTrue(delete_call.called)



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

import sqlalchemy
from sqlalchemy.engine.url import make_url
from sqlalchemy.orm import sessionmaker

from a10_neutron_lbaas.tests.db import session
from a10_neutron_lbaas.tests import test_case


class DbTestBase(test_case.TestCase):

    def setUp(self):
        self._undo = []
        url = os.getenv("PIFPAF_URL")
        if not url:
            self.skipTest("No database URL set")

        sa_url = make_url(url)

        if sa_url.drivername == 'mysql':
            sa_url.drivername = 'mysql+pymysql'

        initial_engine = sqlalchemy.create_engine(str(sa_url))
        initial_connection = initial_engine.connect()
        self._undo.append(initial_connection.close)
        initial_connection.execute("CREATE DATABASE a10_test_db")
        self._undo.append(lambda: initial_engine.execute("DROP DATABASE a10_test_db"))

        sa_url.database = 'a10_test_db'
        self.engine = sqlalchemy.create_engine(str(sa_url))
        self.connection = self.engine.connect()
        self._undo.append(self.connection.close)

    def tearDownDb(self):
        while (self._undo):
            self._undo.pop()()

    def tearDown(self):
        self.tearDownDb()


class UnitTestBase(DbTestBase):

    def setUp(self):
        super(UnitTestBase, self).setUp()

        try:
            session.create_tables(self.connection)
        except Exception as e:
            # tearDown doesn't run if setUp throws an exception!
            self.tearDownDb()
            raise e

        Session = sessionmaker(bind=self.connection)
        self.open_session = Session



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from a10_neutron_lbaas.acos import openstack_mappings as target
from a10_neutron_lbaas.tests.unit import test_base

import mock


class TestA10Openstack(test_base.UnitTestBase):

    def test_source_ip_v2(self):
        mock_client = mock.Mock()
        mock_client.client.slb.service_group.SOURCE_IP_HASH = 'SOURCE_IP'
        expected = mock_client.client.slb.service_group.SOURCE_IP_HASH
        actual = target.service_group_lb_method(mock_client, 'SOURCE_IP')
        self.assertEqual(expected, actual)



# Copyright 2015,  A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy.ext.declarative import clsregistry

from a10_neutron_lbaas.db import api as db_api

# Import models to be created by create_all
import a10_neutron_lbaas.db.models
import a10_neutron_lbaas.db.models.scaling_group

# Suppress pep8 warnings for unused imports
assert a10_neutron_lbaas.db.models
assert a10_neutron_lbaas.db.models.scaling_group

Base = db_api.get_base()


def create_tables(connection, tables=None):
    Base.metadata.create_all(connection, tables=tables)
    return connection


def a10_models():
    return [model for model in Base._decl_class_registry.values()
            if not isinstance(model, clsregistry._ModuleMarker)]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import with_statement
from alembic import context
from sqlalchemy import engine_from_config, pool
from logging.config import fileConfig

from a10_neutron_lbaas import a10_config
from a10_neutron_lbaas import a10_exceptions as ex

VERSION_TABLE = 'alembic_version_a10'

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# Override db location
if getattr(config, 'connection', None) is None:
    a10_cfg = a10_config.A10Config()
    if not a10_cfg.get('use_database'):
        raise ex.InternalError("database not enabled")
    config.set_main_option("sqlalchemy.url", a10_cfg.get('database_connection'))

# Interpret the config file for Python logging.
# This line sets up loggers basically.
fileConfig(config.config_file_name)

# add your model's MetaData object here
# for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
target_metadata = None

# other values from the config, defined by the needs of env.py,
# can be acquired:
# my_important_option = config.get_main_option("my_important_option")
# ... etc.


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    url = config.get_main_option("sqlalchemy.url")
    context.configure(url=url, target_metadata=target_metadata)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = getattr(config, 'connection', None)

    if connectable is None:
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix='sqlalchemy.',
            poolclass=pool.NullPool)

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=VERSION_TABLE
        )

        with context.begin_transaction():
            context.run_migrations()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fake_objs
import test_base


class TestPools(test_base.UnitTestBase):

    def test_create(self):
        methods = {
            'ROUND_ROBIN':
                self.a.last_client.slb.service_group.ROUND_ROBIN,
            'LEAST_CONNECTIONS':
                self.a.last_client.slb.service_group.LEAST_CONNECTION,
            'SOURCE_IP':
                self.a.last_client.slb.service_group.WEIGHTED_LEAST_CONNECTION,
        }
        protocols = {
            'TCP': self.a.last_client.slb.service_group.TCP,
            'UDP': self.a.last_client.slb.service_group.UDP,
        }

        for p in protocols.keys():
            for m in methods.keys():
                self.a.reset_mocks()
                saw_exception = False

                pool = fake_objs.FakePool(p, m)
                self.a.pool.create(None, pool)

                self.print_mocks()
                (self.a.last_client.slb.service_group.create.
                    assert_called())

                if not saw_exception:
                    n = str(self.a.last_client.mock_calls).index(
                        'slb.service_group.create')
                    self.assertTrue(n >= 0)

    def test_update(self):
        old_pool = fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS')
        pool = fake_objs.FakePool('TCP', 'ROUND_ROBIN')
        self.a.pool.update(None, old_pool, pool)
        self.print_mocks()
        self.a.last_client.slb.service_group.update.assert_called()

    def _test_delete(self, pool):
        self.a.pool.delete(None, pool)
        self.print_mocks()

    def test_delete(self):
        pool = fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS')
        pool.members = [fake_objs.FakeMember()]
        pool.health_monitors_status = [{'monitor_id': 'hm1', "pools": [pool]}]
        self.a.pool.neutron.openstack_driver._pool_get_hm.return_value = fake_objs.FakeHM()

        self._test_delete(pool)

        (self.a.last_client.slb.service_group.delete.
            assert_called_with(pool['id']))

    def test_delete_with_hm_dissociates_hm(self):
        pool = fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS')
        hm = fake_objs.FakeHM("TCP")
        hm.pools.append(fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS'))
        pool.members = [fake_objs.FakeMember()]
        pool.health_monitors_status = [{'monitor_id': 'hm1', "pools": [pool]}]

        self.a.pool.neutron.openstack_driver._pool_get_hm.return_value = hm

        self._test_delete(pool)
        self.a.last_client.slb.service_group.update.assert_called_with(
            "id1",
            health_monitor="",
            health_check_disable=True)

    def test_delete_without_health_monitor(self):
        pool = fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS')
        pool.members = [fake_objs.FakePool()]
        pool.health_monitors_status = []
        self._test_delete(pool)
        (self.a.last_client.slb.service_group.delete.
            assert_called_with(pool.id))

    def test_delete_removes_monitor(self):
        pool = fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS')
        pool.members = [fake_objs.FakeMember()]
        pool.health_monitors_status = [{'monitor_id': "hm1"}]
        self.a.pool.delete(None, pool)
        self.a.last_client.slb.hm.delete.assert_called()

    def test_stats(self):
        pool = fake_objs.FakePool('TCP', 'LEAST_CONNECTIONS')
        z = self.a.pool
        z.neutron.pool_get_tenant_id = lambda x, y: 'hello'
        z._get_vip_id = lambda x, y: '2.2.2.2'
        z.stats(None, pool.id)
        self.print_mocks()
        s = str(self.a.last_client.mock_calls)
        self.assertTrue(s.index('slb.virtual_server.stats') >= 0)



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client.errors as acos_errors

from a10_neutron_lbaas.acos import openstack_mappings
from a10_neutron_lbaas import constants
from a10_neutron_lbaas.neutron_ext.db.certificate_db \
    import A10CertificateDbMixin as A10CertificateDb
import a10_neutron_lbaas.v2.wrapper_certmgr as certwrapper
import handler_base_v2
import handler_persist
import v2_context as a10


LOG = logging.getLogger(__name__)


class ListenerHandler(handler_base_v2.HandlerBaseV2):

    def __init__(self, a10_driver, openstack_manager, neutron=None, barbican_client=None,
                 cert_db=None):
        super(ListenerHandler, self).__init__(a10_driver, openstack_manager, neutron)
        self._barbican_client = barbican_client
        self._cert_db = cert_db

    @property
    def cert_db(self):
        if self._cert_db is None:
            self._cert_db = A10CertificateDb()
        return self._cert_db

    @property
    def barbican_client(self):
        if self._barbican_client is None:
            self._barbican_client = certwrapper.CertManagerWrapper(handler=self)
        return self._barbican_client

    def _set(self, set_method, c, context, listener):

        status = c.client.slb.UP
        if not listener.admin_state_up:
            status = c.client.slb.DOWN

        templates = self.meta(listener, "template", {})

        server_args = {}
        cert_data = dict()
        template_args = {}
        protocol = openstack_mappings.vip_protocols(c, listener.protocol)
        binding = None
        os_name = listener.name or None
        # Try Barbican first.  TERMINATED HTTPS requires a default TLS container ID that is
        # checked by the API so we can't fake it out.
        if listener.protocol and listener.protocol == constants.PROTOCOL_TERMINATED_HTTPS:
            if self._set_terminated_https_values(listener, c, cert_data):
                templates["client_ssl"] = {}
                template_name = str(cert_data.get('template_name') or '')
                key_passphrase = str(cert_data.get('key_pass') or '')
                cert_filename = str(cert_data.get('cert_filename') or '')
                key_filename = str(cert_data.get('key_filename') or '')
            else:
                LOG.error("Could not created terminated HTTPS endpoint.")
        # Else, set it up as an HTTP endpoint and attach the A10 cert data.
        elif (self.a10_driver.config.get('use_database') and listener.protocol and
              listener.protocol in [constants.PROTOCOL_HTTPS,
                                    constants.PROTOCOL_HTTP,
                                    constants.PROTOCOL_TCP]):
            try:
                binding = self.cert_db.get_binding_for_listener(context, listener.id)
            except Exception as ex:
                LOG.exception(ex)

            if binding:
                # if the binding is being deleted and the listener wasn't created as https,
                # remove the https port to make room for the original port
                if binding.status == constants.STATUS_DELETING:

                    if (protocol != c.client.slb.virtual_server.vport.HTTPS):
                        self._delete_listener(c, context, listener,
                                              c.client.slb.virtual_server.vport.HTTPS)
                        set_method = c.client.slb.virtual_server.vport.create
                elif self._set_a10_https_values(listener, c, cert_data, binding):
                    # If the binding hasn't been created and the port isn't https
                    # remove the port and re-create it as https.
                    if (binding.status == constants.STATUS_CREATING and
                            protocol != c.client.slb.virtual_server.vport.HTTPS):

                        self._delete_listener(c, context, listener, protocol)
                        set_method = c.client.slb.virtual_server.vport.create

                    protocol = c.client.slb.virtual_server.vport.HTTPS
                    templates["client_ssl"] = {}
                    template_name = str(cert_data.get('template_name') or '')
                    key_passphrase = str(cert_data.get('key_pass') or '')
                    cert_filename = str(cert_data.get('cert_filename') or '')
                    key_filename = str(cert_data.get('key_filename') or '')
                    template_args["template_client_ssl"] = template_name

        if 'client_ssl' in templates:
            template_args["template_client_ssl"] = template_name
            try:
                c.client.slb.template.client_ssl.create(
                    template_name,
                    cert=cert_filename,
                    key=key_filename,
                    passphrase=key_passphrase)
            except acos_errors.Exists:
                c.client.slb.template.client_ssl.update(template_name, cert=cert_filename,
                                                        key=key_filename, passphrase=key_passphrase)

        if 'server_ssl' in templates:
            server_args = {'server_ssl_template': templates['server_ssl']}
            template_args["template_server_ssl"] = template_name
            try:
                c.client.slb.template.server_ssl.create(
                    template_name,
                    cert_filename,
                    key_filename,
                    passphrase=key_passphrase,
                    axapi_args=server_args)
            except acos_errors.Exists:
                c.client.slb.template.server_ssl.update(template_name,
                                                        cert_filename,
                                                        key_filename,
                                                        passphrase=key_passphrase,
                                                        axapi_args=server_args)

        try:
            pool_name = self._pool_name(context, pool_id=listener.default_pool_id)
        except Exception:
            pool_name = None

        persistence = handler_persist.PersistHandler(
            c, context, listener.default_pool)

        # This doesn't do anything anymore.
        vport_meta = self.meta(listener.loadbalancer, 'vip_port', {})
        template_args.update(**self._get_vport_defaults(c, os_name))

        try:
            set_method(
                self.a10_driver.loadbalancer._name(listener.loadbalancer),
                self._meta_name(listener),
                protocol=protocol,
                port=listener.protocol_port,
                service_group_name=pool_name,
                s_pers_name=persistence.s_persistence(),
                c_pers_name=persistence.c_persistence(),
                status=status,
                autosnat=c.device_cfg.get('autosnat'),
                ipinip=c.device_cfg.get('ipinip'),
                source_nat_pool=c.device_cfg.get('source_nat_pool'),
                # Device-level defaults
                vport_defaults=self._get_vport_defaults(c, os_name),
                axapi_body=vport_meta,
                **template_args)
        except acos_errors.Exists:
            pass

    def _set_terminated_https_values(self, listener, c, cert_data):
        is_success = False
        container = None

        c_id = listener.default_tls_container_id if listener.default_tls_container_id else None

        # if there's a barbican container ID, check there.
        if c_id:
            try:
                container = self.barbican_client.get_certificate(c_id, check_only=True,
                                                                 project_id=c.tenant_id)
            except Exception as ex:
                container = None
                LOG.error("Exception encountered retrieving TLS Container %s" % c_id)
                LOG.exception(ex)

            if container:
                base_name = getattr(
                    getattr(container, '_cert_container', None), 'name', '') or c_id

                cert_data["cert_content"] = container.get_certificate()
                cert_data["key_content"] = container.get_private_key()
                cert_data["key_pass"] = container.get_private_key_passphrase()

                cert_data["template_name"] = listener.id

                cert_data["cert_filename"] = "{0}cert.pem".format(base_name)
                cert_data["key_filename"] = "{0}key.pem".format(base_name)

                self._acos_create_or_update(c.client.file.ssl_cert,
                                            file=cert_data["cert_filename"],
                                            cert=cert_data["cert_content"],
                                            size=len(cert_data["cert_content"]),
                                            action="import", certificate_type="pem")

                self._acos_create_or_update(c.client.file.ssl_key,
                                            file=cert_data["key_filename"],
                                            cert=cert_data["key_content"],
                                            size=len(cert_data["key_content"]),
                                            action="import")
                is_success = True
        else:
            LOG.error("default_tls_container_id unspecified for listener.")
        return is_success

    def _set_a10_https_values(self, listener, c, cert_data, binding):
        LOG.info("default_tls_container_id unspecified for listener. Checking A10 DB.")
        is_success = False

        cert_data["cert_content"] = binding.certificate.cert_data or None
        cert_data["key_content"] = binding.certificate.key_data or None
        cert_data["key_pass"] = binding.certificate.password or None

        if len(cert_data["cert_content"]) > 1:
            base_name = listener.id
            cert_data["template_name"] = base_name

            cert_data["cert_filename"] = "{0}cert.pem".format(base_name)

            self._acos_create_or_update(c.client.file.ssl_cert,
                                        file=cert_data["cert_filename"],
                                        cert=cert_data["cert_content"],
                                        size=len(cert_data["cert_content"]),
                                        action="import", certificate_type="pem")

            if len(cert_data.get("key_content") or "") > 0:
                cert_data["key_filename"] = "{0}key.pem".format(base_name)
                self._acos_create_or_update(c.client.file.ssl_key,
                                            file=cert_data["key_filename"],
                                            cert=cert_data["key_content"],
                                            size=len(cert_data["key_content"]),
                                            action="import")

            is_success = True

        return is_success

    def _acos_create_or_update(self, acos_obj, **kwargs):
        try:
            acos_obj.create(**kwargs)
        except acos_errors.Exists:
            acos_obj.update(**kwargs)

    def _create(self, c, context, listener):
        self._set(c.client.slb.virtual_server.vport.create,
                  c, context, listener)

    def create(self, context, listener):
        with a10.A10WriteStatusContext(self, context, listener) as c:
            self._create(c, context, listener)

    def _update(self, c, context, listener):
        self._set(c.client.slb.virtual_server.vport.update, c, context, listener)

    def update(self, context, old_listener, listener):
        with a10.A10WriteStatusContext(self, context, listener) as c:
            self._update(c, context, listener)

    def _delete_listener(self, c, context, listener, protocol):
        try:
            c.client.slb.virtual_server.vport.delete(
                self.a10_driver.loadbalancer._name(listener.loadbalancer),
                self._meta_name(listener),
                protocol=protocol,
                port=listener.protocol_port)
        except acos_errors.NotFound:
            pass

    def _delete(self, c, context, listener):
        # First, remove any existing cert bindings and set the correct protocol for delete.
        # Existence of bindings means the vport has been re-created as https.
        protocol = openstack_mappings.vip_protocols(c, listener.protocol)

        if self.a10_driver.config.use_database and self._remove_existing_bindings(
                c, context, listener):
                protocol = c.client.slb.virtual_server.vport.HTTPS

        # Regular delete, use regular protocol mapping.
        self._delete_listener(
            c, context, listener, protocol)

        # clean up ssl template
        if listener.protocol and listener.protocol == constants.PROTOCOL_TERMINATED_HTTPS:
            try:
                c.client.slb.template.client_ssl.delete(listener.id)
            except acos_errors.NotFound:
                pass

    def delete(self, context, listener):
        with a10.A10DeleteContext(self, context, listener) as c:
            self._delete(c, context, listener)

    def _remove_existing_bindings(self, c, context, listener):
        binding = None

        try:
            binding = self.cert_db.get_binding_for_listener(context, listener.id)
        except Exception as ex:
            LOG.exception(ex)
        # if we have bindings, remove them
        if binding is not None:
            try:
                self.cert_db.delete_a10_certificate_binding(context, binding.id)
            except Exception as ex:
                LOG.exception(ex)
        return binding

    def _get_global_vport_defaults(self, c):
        return c.a10_driver.config.get_vport_defaults()

    def _get_device_vport_defaults(self, c):
        return c.device_cfg.get("vport_defaults")

    def _get_vport_defaults(self, c, vport_name):
        rv = {}
        # Device-specific defaults have precedence over global
        rv.update(self._get_global_vport_defaults(c))
        rv.update(self._get_device_vport_defaults(c))
        if vport_name and len(vport_name) > 0:
            self._get_name_matches(rv, vport_name, self._get_expressions(c))
        return rv

    def _get_expressions(self, c):
        rv = {}
        rv = c.a10_driver.config.get_vport_expressions()
        return rv



# Copyright 2014, Doug Wiegley (dougwig), A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import logging

import acos_client

import a10_config
import monkey_patch
import version

import v1.handler_hm
import v1.handler_member
import v1.handler_pool
import v1.handler_vip

import v2.handler_hm
import v2.handler_lb
import v2.handler_listener
import v2.handler_member
import v2.handler_pool

logging.basicConfig()
LOG = logging.getLogger(__name__)


class A10OpenstackLBBase(object):

    def __init__(self, openstack_driver,
                 plumbing_hooks_class=None,
                 neutron_hooks_module=None,
                 barbican_client=None,
                 config=None,
                 config_dir=None,
                 provider=None,
                 cert_db=None):
        self.openstack_driver = openstack_driver
        self.plumbing_hooks_class = plumbing_hooks_class
        self.neutron = neutron_hooks_module
        self.barbican_client = barbican_client
        self.cert_db = cert_db
        self.config = config
        self.config_dir = config_dir
        self.provider = provider
        self.hooks = None

        LOG.info("A10-neutron-lbaas: pre-initializing, version=%s, acos_client=%s",
                 version.VERSION, acos_client.VERSION)

        monkey = monkey_patch.MonkeyPatch(self.openstack_driver.plugin)
        self.openstack_driver.plugin.stats = monkey.stats

        if provider is not None:
            self._late_init(provider)

    def _late_init(self, provider):
        LOG.info("A10-neutron-lbaas: initializing, version=%s, acos_client=%s, provider=%s",
                 version.VERSION, acos_client.VERSION, provider)

        self.provider = provider
        if self.config is None:
            self.config = a10_config.A10Config(config_dir=self.config_dir, provider=provider)

        if self.plumbing_hooks_class is not None:
            self.hooks = self.plumbing_hooks_class(self)
        else:
            self.hooks = self.config.get('plumbing_hooks_class')(self)

        if self.config.get('verify_appliances'):
            self._verify_appliances()

    def _select_a10_device(self, tenant_id, a10_context=None, lbaas_obj=None, **kwargs):
        if hasattr(self.hooks, 'select_device_with_lbaas_obj'):
            return self.hooks.select_device_with_lbaas_obj(
                tenant_id, a10_context=a10_context, lbaas_obj=lbaas_obj, **kwargs)
        else:
            return self.hooks.select_device(tenant_id, **kwargs)

    def _get_a10_client(self, device_info, **kwargs):
        if hasattr(self.hooks, 'get_a10_client'):
            return self.hooks.get_a10_client(device_info, **kwargs)
        else:
            return acos_client.Client(
                device_info['host'], device_info['api_version'],
                device_info['username'], device_info['password'],
                port=device_info['port'], protocol=device_info['protocol'])

    def _verify_appliances(self):
        LOG.info("A10Driver: verifying appliances")

        if len(self.config.get_devices()) == 0:
            LOG.error("A10Driver: no configured appliances")

        for k, v in self.config.get_devices().items():
            try:
                LOG.info("A10Driver: appliance(%s) = %s", k,
                         self._get_a10_client(v).system.information())
            except Exception:
                LOG.error("A10Driver: unable to connect to configured"
                          "appliance, name=%s", k)


class A10OpenstackLBV2(A10OpenstackLBBase):

    @property
    def lb(self):
        return v2.handler_lb.LoadbalancerHandler(
            self,
            self.openstack_driver.load_balancer,
            neutron=self.neutron)

    @property
    def loadbalancer(self):
        return self.lb

    @property
    def listener(self):
        return v2.handler_listener.ListenerHandler(
            self,
            self.openstack_driver.listener,
            neutron=self.neutron,
            barbican_client=self.barbican_client,
            cert_db=self.cert_db)

    @property
    def pool(self):
        return v2.handler_pool.PoolHandler(
            self, self.openstack_driver.pool,
            neutron=self.neutron)

    @property
    def member(self):
        return v2.handler_member.MemberHandler(
            self,
            self.openstack_driver.member,
            neutron=self.neutron)

    @property
    def hm(self):
        return v2.handler_hm.HealthMonitorHandler(
            self,
            self.openstack_driver.health_monitor,
            neutron=self.neutron)


class A10OpenstackLBV1(A10OpenstackLBBase):

    @property
    def pool(self):
        return v1.handler_pool.PoolHandler(self)

    @property
    def vip(self):
        return v1.handler_vip.VipHandler(self)

    @property
    def member(self):
        return v1.handler_member.MemberHandler(self)

    @property
    def hm(self):
        return v1.handler_hm.HealthMonitorHandler(self)



# Copyright 2017, A10 Networks
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from neutron_lbaas.common.cert_manager import cert_manager

from a10_neutron_lbaas.tests import test_case
from a10_neutron_lbaas.v2 import wrapper_certmgr


class TestCertManagerWrapper(test_case.TestCase):

    def setUp(self, **kwargs):
        super(TestCertManagerWrapper, self).setUp(**kwargs)

        certmgr = mock.create_autospec(cert_manager.CertManager)
        self.target = wrapper_certmgr.CertManagerWrapper(certmgr=certmgr)

    def test_get_certificate_type_checks(self):
        self.target.get_certificate('fake-container-id',
                                    project_id='fake-project-id',
                                    check_only=True)

    def test_store_cert_type_checks(self):
        self.target.store_cert('fake-certificate',
                               'fake-private-key',
                               project_id='fake-project-id')

    def test_delete_cert_type_checks(self):
        self.target.delete_cert('fake-certificate-id',
                                project_id='fake-project-id')
